import numpy as np
import pytest

from psyneulink.core import llvm as pnlvm

SEED = 0

@pytest.mark.benchmark(group="Philox integer PRNG")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize('seed, expected', [
    (0, [259491006799949737,  4754966410622352325,  8698845897610382596, 1686395276220330909, 18061843536446043542, 4723914225006068263]),
    (-5, [4936860362606747269, 11611290354192475889, 2015254117581537576, 4620074701282684350, 9574602527017877750, 2811009141214824706]),
    (15, [322160557315224026, 10187298772616605914, 11130303561932346278, 3540317624683947565, 245468466731153020, 17669502083357198575]),
    (0xfeedcafe, [14360762734736817955, 5188080951818105836, 1417692977344505657, 15919241602363537044, 11006348070701344872, 12539562470140893435]),
])
def test_random_int64(benchmark, mode, seed, expected):
    if mode == 'numpy':
        state = np.random.Philox([np.int64(seed).astype(np.uint64)])
        prng = np.random.Generator(state)
        def f():
            # Get uint range [0, MAX] to avoid any intermediate caching of random bits
            return prng.integers(0xffffffffffffffff, dtype=np.uint64, endpoint=True)

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state = init_fun.np_buffer_for_arg(0)
        init_fun(state, seed)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_int64')

        def f():
            out = gen_fun.np_buffer_for_arg(1)
            gen_fun(state, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)

        init_fun.cuda_call(gpu_state, np.int64(seed))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_int64')
        out = gen_fun.np_buffer_for_arg(1)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    # Get >4 samples to force regeneration of Philox buffer
    res = [f(), f(), f(), f(), f(), f()]
    np.testing.assert_array_equal(res, expected)
    benchmark(f)


@pytest.mark.benchmark(group="Philox integer PRNG")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize("bounds, expected",
    [((0xffffffff,), [582496168, 60417457, 4027530180, 1107101888, 1659784451, 2025357888]),
     ((15,), [2, 0, 14, 3, 5, 7]),
     ((0,15), [2, 0, 14, 3, 5, 7]),
     ((5,0xffff), [8892, 926, 61454, 16896, 25328, 30906]),
    ], ids=lambda x: str(x) if len(x) != 6 else "")
def test_random_int32_bounded(benchmark, mode, bounds, expected):
    if mode == 'numpy':
        state = np.random.Philox([SEED])
        prng = np.random.Generator(state)

        def f():
            return prng.integers(*bounds, dtype=np.uint32, endpoint=False)

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state = init_fun.np_buffer_for_arg(0)
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_int32_bounded')

        def f():
            lower, upper = bounds if len(bounds) == 2 else (0, bounds[0])
            out = gen_fun.np_buffer_for_arg(3)
            gen_fun(state, lower, upper, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)
        init_fun.cuda_call(gpu_state, np.int64(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_int32_bounded')
        out = gen_fun.np_buffer_for_arg(3)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            lower, upper = bounds if len(bounds) == 2 else (0, bounds[0])
            gen_fun.cuda_call(gpu_state, np.uint32(lower), np.uint32(upper), gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    # Get >4 samples to force regeneration of Philox buffer
    res = [f(), f(), f(), f(), f(), f()]
    np.testing.assert_array_equal(res, expected)
    benchmark(f)

@pytest.mark.benchmark(group="Philox integer PRNG")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
def test_random_int32(benchmark, mode):
    if mode == 'numpy':
        state = np.random.Philox([SEED])
        prng = np.random.Generator(state)\

        def f():
            # Get uint range [0, MAX] to avoid any intermediate caching of random bits
            return prng.integers(0xffffffff, dtype=np.uint32, endpoint=True)

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state = init_fun.np_buffer_for_arg(0)
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_int32')

        def f():
            out = gen_fun.np_buffer_for_arg(1)
            gen_fun(state, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)
        init_fun.cuda_call(gpu_state, np.int64(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_int32')
        out = gen_fun.np_buffer_for_arg(1)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    # Get >4 samples to force regeneration of Philox buffer
    res = [f(), f(), f(), f(), f(), f()]
    np.testing.assert_array_equal(res, [582496169, 60417458, 4027530181, 1107101889, 1659784452, 2025357889])
    benchmark(f)


@pytest.mark.benchmark(group="Philox floating point PRNG")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
def test_random_double(benchmark, mode):
    if mode == 'numpy':
        state = np.random.Philox([SEED])
        prng = np.random.Generator(state)

        def f():
            return prng.random(dtype=np.float64)

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state = init_fun.np_buffer_for_arg(0)
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_double')

        def f():
            out = gen_fun.np_buffer_for_arg(1)
            gen_fun(state, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)
        init_fun.cuda_call(gpu_state, np.int64(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_double')
        out = gen_fun.np_buffer_for_arg(1)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    res = [f(), f()]
    np.testing.assert_allclose(res, [0.014067035665647709, 0.2577672456246177])
    benchmark(f)


@pytest.mark.benchmark(group="Philox floating point PRNG")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
def test_random_float(benchmark, mode):
    if mode == 'numpy':
        state = np.random.Philox([SEED])
        prng = np.random.Generator(state)

        def f():
            return prng.random(dtype=np.float32)

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state = init_fun.np_buffer_for_arg(0)
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_float')

        def f():
            out = gen_fun.np_buffer_for_arg(1)
            gen_fun(state, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)
        init_fun.cuda_call(gpu_state, np.int64(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_float')
        out = gen_fun.np_buffer_for_arg(1)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    res = [f(), f()]
    np.testing.assert_allclose(res, [0.13562285900115967, 0.014066934585571289], rtol=1e-5, atol=1e-8)
    benchmark(f)


@pytest.mark.benchmark(group="Philox Normal distribution")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize('fp_type', [pnlvm.ir.DoubleType(), pnlvm.ir.FloatType()], ids=str)
def test_random_normal(benchmark, mode, fp_type):
    if mode != 'numpy':
        # Instantiate builder context with the desired type
        pnlvm.LLVMBuilderContext(fp_type)

    dtype = np.float64 if fp_type is pnlvm.ir.DoubleType() else np.float32
    if mode == 'numpy':
        state = np.random.Philox([SEED])
        prng = np.random.Generator(state)

        def f():
            return prng.standard_normal(dtype=dtype)

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state = init_fun.np_buffer_for_arg(0)
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_normal')

        def f():
            out = gen_fun.np_buffer_for_arg(1)
            gen_fun(state, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)
        init_fun.cuda_call(gpu_state, np.int64(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_normal')
        out = gen_fun.np_buffer_for_arg(1)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    res = [f() for i in range(191000)]
    if fp_type is pnlvm.ir.DoubleType():
        np.testing.assert_allclose(res[0:2], [-0.2059740286292238, -0.12884495093462758])
        # 208 doesn't take the fast path but wraps around the main loop
        np.testing.assert_allclose(res[207:211],
                                   [-0.768690647997579, 0.4301874289485477, -0.7803640491708955, -1.146089287628737])
        # 450 doesn't take the fast path or wrap around the main loop,
        # but takes the special condition at the end of the loop
        np.testing.assert_allclose(res[449:453],
                                   [-0.7713655663874537, -0.5638348710823825, -0.9415838853097869, 0.6212784278881248])
        # 2013 takes the rare secondary loop and exists in the first iteration
        # taking the positive value
        np.testing.assert_allclose(res[2011:2015],
                                   [0.4201922976982861, 2.7021541445373916, 3.7809967764329375, 0.19919094793393655])
        # 5136 takes the rare secondary loop and exists in the first iteration
        # taking the negative value
        np.testing.assert_allclose(res[5134:5138],
                                   [0.12317411414687844, -0.17846827974421134, -3.6579887696059714, 0.2501530374224693])
        # 190855 takes the rare secondary loop and needs more than one iteration
        np.testing.assert_allclose(res[190853:190857],
                                   [-0.26418319904491194, 0.35889007879353746, -3.843811523424439, -1.5256469840469997])

    elif fp_type is pnlvm.ir.FloatType():
        # The indices are taken from above and don't have special meaning.
        np.testing.assert_allclose(res[0:2], [-0.24822916090488434, -0.02676701545715332])
        np.testing.assert_allclose(res[207:211],
                                   [-0.33086925745010376, -1.024695873260498, -0.5162619352340698, -0.15033885836601257])
        np.testing.assert_allclose(res[449:453],
                                   [-0.2223609834909439, 0.16769859194755554, -0.7806711196899414, 0.5867824554443359])
        np.testing.assert_allclose(res[2011:2015],
                                   [0.1979091316461563, -0.23467595875263214, 1.1458240747451782, -1.0285860300064087])
        np.testing.assert_allclose(res[5134:5138],
                                   [-1.0523858070373535, -3.007537603378296, -0.4331461489200592, -0.8841480612754822])
        np.testing.assert_allclose(res[190853:190857],
                                   [-0.8958197236061096, 0.10532315075397491, 2.000257730484009, -1.129721999168396])

    assert not any(np.isnan(res)), list(np.isnan(res)).index(True)
    benchmark(f)

@pytest.mark.benchmark(group="Philox Binomial distribution")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize('fp_type', [pnlvm.ir.DoubleType(), pnlvm.ir.FloatType()],
                         ids=str)
@pytest.mark.parametrize('n', [1])
@pytest.mark.parametrize('p, exp_64, exp_32', [
    (0, [0], [0]),
    (0.1, [0xa0c0100], [0x20440250]),
    (0.33, [0xa2c8186], [0x20440650]),
    (0.5, [0xa2c81c6], [0x226c8650]),
    (0.66, [0xf5d37e79], [0xdfbbf9af]),
    (0.95, [0xf7f3ffff], [0xffbffdaf]),
    (1, [0xffffffff], [0xffffffff]),
    ])
def test_random_binomial(benchmark, mode, fp_type, n, p, exp_64, exp_32):
    if mode != 'numpy':
        # Instantiate builder context with the desired type
        pnlvm.LLVMBuilderContext(fp_type)

    # numpy always uses fp64 uniform sampling
    exp = exp_64 if fp_type is pnlvm.ir.DoubleType() or mode == 'numpy' else exp_32
    if mode == 'numpy':
        state = np.random.Philox([SEED])
        prng = np.random.Generator(state)

        def f():
            return prng.binomial(n, p)

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state = init_fun.np_buffer_for_arg(0)
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_binomial')
        n = np.asarray(n, dtype=gen_fun.np_arg_dtypes[1])
        p = np.asarray(p, dtype=gen_fun.np_arg_dtypes[2])

        def f():
            out = gen_fun.np_buffer_for_arg(1)
            gen_fun(state, n, p, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_init')
        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)
        init_fun.cuda_call(gpu_state, np.int64(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_philox_rand_binomial')
        gpu_n = pnlvm.jit_engine.pycuda.driver.In(np.asarray(n, dtype=gen_fun.np_arg_dtypes[1]))
        gpu_p = pnlvm.jit_engine.pycuda.driver.In(np.asarray(p, dtype=gen_fun.np_arg_dtypes[2]))
        out = gen_fun.np_buffer_for_arg(1)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            gen_fun.cuda_call(gpu_state, gpu_n, gpu_p, gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    res = [f() for i in range(32)]
    res = int(''.join(str(x) for x in res), 2)
    assert res == exp[n - 1]

    benchmark(f)



import graph_scheduler
import modeci_mdf.functions.standard

import psyneulink as pnl

A = pnl.TransferMechanism(name='A')
B = pnl.ProcessingMechanism(
    name='B',
    function=pnl.UserDefinedFunction(
        modeci_mdf.functions.standard.mdf_functions['sin']['function'],
        scale=1
    )
)
C = pnl.ProcessingMechanism(
    name='C',
    function=pnl.UserDefinedFunction(
        modeci_mdf.functions.standard.mdf_functions['cos']['function'],
        scale=1
    )
)
comp = pnl.Composition(name='comp', pathways=[A, B, C])
comp.scheduler.add_condition(B, pnl.EveryNCalls(A, 2))
comp.scheduler.add_condition(C, graph_scheduler.EveryNCalls(B, 2))



import math

import numpy as np
import pytest

from psyneulink.core.components.functions.nonstateful.transferfunctions import Linear, Logistic
from psyneulink.core.components.functions.userdefinedfunction import UserDefinedFunction
from psyneulink.core.components.mechanisms.processing import ProcessingMechanism
from psyneulink.core.components.mechanisms.processing import TransferMechanism
from psyneulink.core.compositions.composition import Composition


# default val is same shape as expected output
# we only use param1 and param2 to avoid automatic shape changes of the variable
def binAdd(_, param1, param2):
    return param1 + param2


def binSub(_, param1, param2):
    return param1 - param2


def binMul(_, param1, param2):
    return param1 * param2


def binDiv(_, param1, param2):
    return param1 / param2


def binPow(_, param1, param2):
    return param1 ** param2


@pytest.mark.parametrize("param1", [1, np.ones(2), np.ones((2, 2))], ids=['scalar', 'vector', 'matrix'])
@pytest.mark.parametrize("param2", [2, np.ones(2) * 2, np.ones((2, 2)) * 2], ids=['scalar', 'vector', 'matrix'])
@pytest.mark.parametrize("func", [binAdd, binSub, binMul, binDiv, binPow])
@pytest.mark.benchmark(group="Function UDF")
def test_user_def_bin_arith(param1, param2, func, func_mode, benchmark):

    U = UserDefinedFunction(custom_function=func, param1=param1, param2=param2)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, 0)
    np.testing.assert_allclose(val, func(0, param1=param1, param2=param2))


def unNot(variable):
    var1 = False
    # compiled UDFs don't support python bool type outputs
    if not var1:
        return 0.0
    else:
        return 1.0


def binAnd(variable):
    var1 = True
    var2 = False
    # compiled UDFs don't support python bool type outputs
    if var1 and var2:
        return 0.0
    else:
        return 1.0

def triAnd(variable):
    var1 = True
    var2 = False
    # compiled UDFs don't support python bool type outputs
    if var1 and var2 and False:
        return 0.0
    else:
        return 1.0


def binOr(variable):
    var1 = True
    var2 = False
    # compiled UDFs don't support python bool type outputs
    if var1 or var2:
        return 1.0
    else:
        return 0.0


def triOr(variable):
    var1 = True
    var2 = False
    # compiled UDFs don't support python bool type outputs
    if var1 or var2 or True:
        return 1.0
    else:
        return 0.0


def multiAndOr(variable):
    var1 = True
    var2 = False
    # compiled UDFs don't support python bool type outputs
    if var1 or var2 and True or True and var1:
        return 1.0
    else:
        return 0.0


def varAnd(var):
    return var[0] and var[1]


def varOr(var):
    return var[0] or var[1]


def varNot(var):
    # compiled UDFs don't support python bool type outputs
    if not var[0]:
        return 0.0
    else:
        return 1.0


@pytest.mark.parametrize("op,var, expected", [
    (unNot, 0, 0.0),
    (binAnd, 0, 1.0),
    (binOr, 0, 1.0),
    (triAnd, 0, 1.0),
    (triOr, 0, 1.0),
    (multiAndOr, 0, 1.0),
    (varAnd, [0.0, 0.0], 0.0),
    (varAnd, [5.0, -0.0], -0.0),
    (varAnd, [0.0, -2.0], 0.0),
    (varAnd, [5.0, -2.0], -2.0),
    (varOr, [-0.0, 0.0], 0.0),
    (varOr, [3.0, 0.0], 3.0),
    (varOr, [0.0, -3.0], -3.0),
    (varOr, [1.5, -1.0], 1.5),
    (varNot, [1.5], 1.0),
    (varNot, [0.0], 0.0),
    ])
@pytest.mark.benchmark(group="Function UDF")
def test_user_def_func_boolop(op, var, expected, func_mode, benchmark):
    U = UserDefinedFunction(custom_function=op, default_variable=var)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, var)
    assert val == expected


def binEQ(variable, var1, var2):
    if var1 == var2:
        return 1.0
    else:
        return 0.0


def binNE(variable, var1, var2):
    if var1 != var2:
        return 1.0
    else:
        return 0.0


def binLT(variable, var1, var2):
    if var1 < var2:
        return 1.0
    else:
        return 0.0


def binLE(variable, var1, var2):
    if var1 <= var2:
        return 1.0
    else:
        return 0.0


def binGT(variable, var1, var2):
    if var1 > var2:
        return 1.0
    else:
        return 0.0


def binGE(variable, var1, var2):
    if var1 >= var2:
        return 1.0
    else:
        return 0.0

@pytest.mark.parametrize("func,var1,var2,expected", [
                    (binEQ, 1.0, 2.0, 0.0),
                    (binNE, 1.0, 2.0, 1.0),
                    (binLT, 1.0, 2.0, 1.0),
                    (binLE, 1.0, 2.0, 1.0),
                    (binGT, 1.0, 2.0, 0.0),
                    (binGE, 1.0, 2.0, 0.0),
                    ])
@pytest.mark.benchmark(group="Function UDF")
def test_user_def_func_cmpop(func, var1, var2, expected, func_mode, benchmark):
    U = UserDefinedFunction(custom_function=func, default_variable=[0], var1=var1, var2=var2)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, [0])
    np.testing.assert_allclose(expected, val)

@pytest.mark.parametrize("op,var1,var2,expected", [ # parameter is string since compiled udf doesn't support closures as of present
                    ("Eq", 1.0, 2.0, 0.0),
                    ("Eq", [1.0, 2.0], [1.0, 2.0], [1.0, 1.0]),
                    ("Eq", 1.0, [1.0, 2.0], [1.0, 0.0]),
                    ("Eq", [2.0, 1.0], 1.0, [0.0, 1.0]),
                    ("Eq", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[1.0, 0.0], [0.0, 0.0]]),
                    ("Eq", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[1.0, 0.0], [0.0, 0.0]]),
                    ("Eq", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[1.0, 1.0], [1.0, 1.0]]),
                    ("NotEq", 1.0, 2.0, 1.0),
                    ("NotEq", [1.0, 2.0], [1.0, 2.0], [0.0, 0.0]),
                    ("NotEq", 1.0, [1.0, 2.0], [0.0, 1.0]),
                    ("NotEq", [2.0, 1.0], 1.0, [1.0, 0.0]),
                    ("NotEq", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[0.0, 1.0], [1.0, 1.0]]),
                    ("NotEq", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[0.0, 1.0], [1.0, 1.0]]),
                    ("NotEq", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[0.0, 0.0], [0.0, 0.0]]),
                    ("Lt", 1.0, 2.0, 1.0),
                    ("Lt", [1.0, 2.0], [1.0, 2.0], [0.0, 0.0]),
                    ("Lt", 1.0, [1.0, 2.0], [0.0, 1.0]),
                    ("Lt", [2.0, 1.0], 1.0, [0.0, 0.0]),
                    ("Lt", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[0.0, 0.0], [0.0, 0.0]]),
                    ("Lt", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[0.0, 1.0], [1.0, 1.0]]),
                    ("Lt", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[0.0, 0.0], [0.0, 0.0]]),
                    ("LtE", 1.0, 2.0, 1.0),
                    ("LtE", [1.0, 2.0], [1.0, 2.0], [1.0, 1.0]),
                    ("LtE", 1.0, [1.0, 2.0], [1.0, 1.0]),
                    ("LtE", [2.0, 1.0], 1.0, [0.0, 1.0]),
                    ("LtE", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[1.0, 0.0], [0.0, 0.0]]),
                    ("LtE", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[1.0, 1.0], [1.0, 1.0]]),
                    ("LtE", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[1.0, 1.0], [1.0, 1.0]]),
                    ("Gt", 1.0, 2.0, 0.0),
                    ("Gt", [1.0, 2.0], [1.0, 2.0], [0.0, 0.0]),
                    ("Gt", 1.0, [1.0, 2.0], [0.0, 0.0]),
                    ("Gt", [2.0, 1.0], 1.0, [1.0, 0.0]),
                    ("Gt", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[0.0, 1.0], [1.0, 1.0]]),
                    ("Gt", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[0.0, 0.0], [0.0, 0.0]]),
                    ("Gt", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[0.0, 0.0], [0.0, 0.0]]),
                    ("GtE", 1.0, 2.0, 0.0),
                    ("GtE", [1.0, 2.0], [1.0, 2.0], [1.0, 1.0]),
                    ("GtE", 1.0, [1.0, 2.0], [1.0, 0.0]),
                    ("GtE", [2.0, 1.0], 1.0, [1.0, 1.0]),
                    ("GtE", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[1.0, 1.0], [1.0, 1.0]]),
                    ("GtE", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[1.0, 0.0], [0.0, 0.0]]),
                    ("GtE", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[1.0, 1.0], [1.0, 1.0]]),
                    ])
@pytest.mark.benchmark(group="Function UDF")
def test_user_def_func_cmpop_numpy(op, var1, var2, expected, func_mode, benchmark):
    # we explicitly use np here to ensure that the result is castable to float in the scalar-scalar case
    if op == "Eq":
        def myFunction(variable, var1, var2):
            return np.equal(var1, var2).astype(float)
    elif op == "NotEq":
        def myFunction(variable, var1, var2):
            return np.not_equal(var1, var2).astype(float)
    elif op == "Lt":
        def myFunction(variable, var1, var2):
            return np.less(var1, var2).astype(float)
    elif op == "LtE":
        def myFunction(variable, var1, var2):
            return np.less_equal(var1, var2).astype(float)
    elif op == "Gt":
        def myFunction(variable, var1, var2):
            return np.greater(var1, var2).astype(float)
    elif op == "GtE":
        def myFunction(variable, var1, var2):
            return np.greater_equal(var1, var2).astype(float)

    U = UserDefinedFunction(custom_function=myFunction, default_variable=[0], var1=var1, var2=var2)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, [0])
    np.testing.assert_allclose(expected, val)


def simpleFun(variable, param1, param2):
    return variable * 2 + param2


def condReturn(variable, param1, param2):
    if variable[0]:
        return param1 + 0.5
    return param2 + 0.3


def condValReturn(variable, param1, param2):
    if variable[0]:
        val = param1 + 0.5
    else:
        val = param2 + 0.3
    return val

def lambdaGen():
    return lambda var, param1, param2: var + param1 * param2


@pytest.mark.parametrize("func,var,params,expected", [
    (simpleFun, [1, 3], {"param1":None, "param2":3}, [5, 9]),
    (condReturn, [0], {"param1":1, "param2":2}, [2.3]),
    (condReturn, [1], {"param1":1, "param2":2}, [1.5]),
    (condValReturn, [0], {"param1":1, "param2":2}, [2.3]),
    (condValReturn, [1], {"param1":1, "param2":2}, [1.5]),
    (lambdaGen(), [3], {"param1":3, "param2":-0.5}, [1.5]),
])
@pytest.mark.benchmark(group="Function UDF")
def test_user_def_func(func, var, params, expected, func_mode, benchmark):

    U = UserDefinedFunction(custom_function=func, default_variable=var, **params)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, var)
    np.testing.assert_allclose(val, expected)


def branchOnVarCmp(variable, param1, param2):
    if variable[0][0] > 0 and variable[0][1] > 0:
        return variable * 2 + param2
    else:
        return variable * -2 + param2


def branchOnVarFloat(variable, param1, param2):
    if variable[0]:
        return 1.0
    else:
        return 0.0


def swap(variable, param1, param2):
    val = variable[0]
    variable[0] = variable[1]
    variable[1] = val
    return variable


def indexLit(variable, param1, param2):
    return [1,2,3,4][variable[0] + 2]


@pytest.mark.parametrize("func,var,expected", [
    (branchOnVarCmp, [[1, 3]], [[5, 9]]),
    (branchOnVarCmp, [[-1, 3]], [[5, -3]]),
    (branchOnVarFloat, [0.0], [0.0]),
    (branchOnVarFloat, [-0.0], [0.0]),
    (branchOnVarFloat, [-1.5], [1.0]),
    (branchOnVarFloat, [1.5], [1.0]),
    (branchOnVarFloat, [float("Inf")], [1.0]),
    (branchOnVarFloat, [float("-Inf")], [1.0]),
    (branchOnVarFloat, [float("NaN")], [1.0]),
    (branchOnVarFloat, [float("-NaN")], [1.0]),
    (swap, [-1, 3], [3, -1]),
    (indexLit, [0], [3]),
])
@pytest.mark.benchmark(group="Function UDF")
def test_user_def_func_branching(func, var, expected, func_mode, benchmark):

    U = UserDefinedFunction(custom_function=func, default_variable=var, param2=3)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, var)
    np.testing.assert_allclose(val, expected)


@pytest.mark.benchmark(group="Function UDF")
def test_user_def_func_variable_index(func_mode, benchmark):
    def myFunction(variable):
        variable[0][0] = variable[0][0] + 5
        variable[0][1] = variable[0][1] + 7
        return variable

    U = UserDefinedFunction(custom_function=myFunction, default_variable=[[0, 0]])
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, [[1, 3]])
    np.testing.assert_allclose(val, [[6, 10]])


def unarySubVar(variable, param):
    return -variable


def unarySubParam(variable, param):
    return -param


def unaryAddVar(variable, param):
    return +variable


def unaryAddParam(variable, param):
    return +param


@pytest.mark.parametrize("variable", [
                    1,
                    np.ones(2),
                    np.ones((2)),
                    np.ones((2, 2))
                    ], ids=["scalar", "vec", "vec-2d", "mat"])
@pytest.mark.parametrize("func", [unarySubVar, unarySubParam, unaryAddVar, unaryAddParam])
@pytest.mark.benchmark(group="Function UDF")
def test_user_def_func_unary(func, variable, func_mode, benchmark):
    U = UserDefinedFunction(custom_function=func, default_variable=variable, param=variable)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, variable)
    np.testing.assert_allclose(val, func(variable, param=variable))


@pytest.mark.benchmark(group="Function UDF")
def test_user_def_reward_func(func_mode, benchmark):
    variable = [[1,2,3,4]]
    def myFunction(x,t0=0.48):
        return (x[0][0]>0).astype(float) * (x[0][2]>0).astype(float) / (np.max([x[0][1],x[0][3]]) + t0)
    U = UserDefinedFunction(custom_function=myFunction, default_variable=variable, param=variable)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, variable)
    np.testing.assert_allclose(val, 0.2232142857142857)


@pytest.mark.parametrize("dtype, expected", [ # parameter is string since compiled udf doesn't support closures as of present
                    ("SCALAR_VAR", 1.0),
                    ("VECTOR_VAR", [1,2]),
                    ("MATRIX_VAR", [[1,2],[3,4]]),
                    ("BOOL_VAR", 1.0),
                    ("TUPLE_VAR", (1, 2, 3, 4)),
                    ("SCALAR_LIT", 1.0),
                    ("VECTOR_LIT", [1,2]),
                    ("MATRIX_LIT", [[1,2],[3,4]]),
                    ("TUPLE_LIT", (1, 2, 3, 4)),
                    ])
@pytest.mark.benchmark(group="Function UDF")
def test_user_def_func_return(dtype, expected, func_mode, benchmark):
    if dtype == "SCALAR_VAR":
        def myFunction(variable):
            var = 1.0
            return var
    elif dtype == "VECTOR_VAR":
        def myFunction(variable):
            var = [1,2]
            return var
    elif dtype == "MATRIX_VAR":
        def myFunction(variable):
            var = [[1,2],[3,4]]
            return var
    elif dtype == "BOOL_VAR":
        def myFunction(variable):
            var = True
            return 1.0
    elif dtype == "TUPLE_VAR":
        def myFunction(variable):
            var = (1, 2, 3, 4)
            return var
    elif dtype == "SCALAR_LIT":
        def myFunction(variable):
            return 1.0
    elif dtype == "VECTOR_LIT":
        def myFunction(variable):
            return [1,2]
    elif dtype == "MATRIX_LIT":
        def myFunction(variable):
            return [[1,2],[3,4]]
    elif dtype == "TUPLE_LIT":
        def myFunction(variable):
            return (1, 2, 3, 4)

    U = UserDefinedFunction(custom_function=myFunction, default_variable=0)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, 0)
    np.testing.assert_allclose(val, expected)


@pytest.mark.parametrize("op,variable,expected", [ # parameter is string since compiled udf doesn't support closures as of present
                ("TANH", [[1, 3]], [[0.76159416, 0.99505475]]),
                ("EXP", [[1, 3]], [[2.71828183, 20.08553692]]),
                ("SQRT", [[1, 3]], [[1.0, 1.7320508075688772]]),
                ("SHAPE", [1, 2], [2]),
                ("SHAPE", [[1, 3]], [1, 2]),
                ("ASTYPE_FLOAT", [1], [1.0]),
                ("ASTYPE_INT", [-1.5], [-1.0]),
                ("NP_MAX", 5.0, 5.0),
                ("NP_MAX", [0.0, 0.0], 0),
                ("NP_MAX", [1.0, 2.0], 2),
                ("NP_MAX", [1.0, 2.0, float("-Inf"), float("Inf"), float("NaN")], float("NaN")),
                ("NP_MAX", [[2.0, 1.0], [6.0, 2.0]], 6),
                ("NP_MAX", [[[-2.0, -1.0], [-6.0, -2.0]],[[2.0, 1.0], [6.0, 2.0]]], 6),
                ("NP_MAX", [[float('-Inf'), 1.0], [6.0, 2.0]], 6),
                ("NP_MAX", [[float('Inf'), 1.0], [6.0, 2.0]], float('Inf')),
                ("NP_MAX", [[float('NaN'), 1.0], [6.0, 2.0]], float('NaN')),
                ("NP_MAX", [[float('-NaN'), 1.0], [6.0, 2.0]], float('-NaN')),
                ("NP_MAX", [[5.0, float('-Inf'), 1.0], [3.0, 6.0, 2.0]], 6),
                ("NP_MAX", [[5.0, float('Inf'), 1.0], [3.0, 6.0, 2.0]], float('Inf')),
                ("NP_MAX", [[5.0, float('NaN'), 1.0], [3.0, 6.0, 2.0]], float('NaN')),
                ("NP_MAX", [[5.0, float('NaN'), 1.0], [3.0, 6.0, 2.0]], float('-NaN')),
                ("FLATTEN", [[1.0, 2.0], [3.0, 4.0]], [1.0, 2.0, 3.0, 4.0])
                ])
@pytest.mark.benchmark(group="Function UDF")
def test_user_def_func_numpy(op, variable, expected, func_mode, benchmark):
    if op == "TANH":
        def myFunction(variable):
            return np.tanh(variable)
    elif op == "EXP":
        def myFunction(variable):
            return np.exp(variable)
    elif op == "SQRT":
        def myFunction(variable):
            return np.sqrt(variable)
    elif op == "SHAPE":
        def myFunction(variable):
            return variable.shape
    elif op == "ASTYPE_FLOAT":
        def myFunction(variable):
            return variable.astype(float)
    elif op == "ASTYPE_INT":
        # return types cannot be integers, so we cast back to float and check for truncation
        def myFunction(variable):
            return variable.astype(int).astype(float)
    elif op == "NP_MAX":
        def myFunction(variable):
            return np.max(variable)
    elif op == "FLATTEN":
        def myFunction(variable):
            return variable.flatten()

    U = UserDefinedFunction(custom_function=myFunction, default_variable=variable)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, variable)
    np.testing.assert_allclose(val, expected, equal_nan=True)


@pytest.mark.benchmark(group="UDF in Mechanism")
def test_udf_in_mechanism(mech_mode, benchmark):
    def myFunction(variable, param1, param2):
        return sum(variable[0]) + 2

    myMech = ProcessingMechanism(function=myFunction, input_shapes=4, name='myMech')
    # assert 'param1' in myMech.parameter_ports.names # <- FIX reinstate when problem with function params is fixed
    # assert 'param2' in myMech.parameter_ports.names # <- FIX reinstate when problem with function params is fixed
    e = pytest.helpers.get_mech_execution(myMech, mech_mode)

    val = benchmark(e, [-1, 2, 3, 4])
    np.testing.assert_allclose(val, [[10]])


@pytest.mark.parametrize("op,variable,expected", [ # parameter is string since compiled udf doesn't support closures as of present
                ("SUM", [1.0, 3.0], 12),
                ("SUM", [[1.0], [3.0]], [12.0]),
                ("SUM", [[1.0, 2.0], [3.0, 4.0]], [12.0, 18.0]),
                ("LEN", [1.0, 3.0], 8),
                ("LEN", [[1.0], [3.0]], 8),
                ("MAX_MULTI", [1.0, 3.0, 2.0], 6),
                ("MAX_TUPLE", [1.0, 3.0, 2.0], 6),
                ("MAX", [1.0, 3.0, 2.0], 3.0),
                ("MAX", [1.0, float("Inf"), 2.0], float("Inf")),
                ("MAX", [1.0, float("NaN"), 2.0], 2.0),
                ("MAX", [1.0, float("-Inf"), 2.0], 2.0),
                ])
@pytest.mark.benchmark(group="Function UDF")
def test_user_def_func_builtin(op, variable, expected, func_mode, benchmark):
    if op == "SUM":
        def myFunction(var):
            return sum(var) + sum((var[0], var[1])) + sum([var[0], var[1]])
    elif op == "LEN":
        def myFunction(var):
            return len(var) + len((var[0], var[1])) + len([var[0], var[1]]) + len((1.0, (1,2)))
    elif op == "MAX":
        def myFunction(variable):
            return max(variable)
    elif op == "MAX_MULTI":
        # special cased, since passing in multiple variables without a closure is hard
        def myFunction(variable):
            return max(variable[0], variable[1], variable[2], -5, 6)
    elif op == "MAX_TUPLE":
        def myFunction(variable):
            return max((variable[0], variable[1], variable[2], -5, 6))

    U = UserDefinedFunction(custom_function=myFunction, default_variable=variable)
    e = pytest.helpers.get_func_execution(U, func_mode)

    val = benchmark(e, variable)
    np.testing.assert_allclose(val, expected)


@pytest.mark.parametrize(
    "func, args, expected",
    [
        (math.fsum, ([1, 2, 3], ), 6),
        (math.sin, (math.pi / 2, ), 1),
    ]
)
@pytest.mark.benchmark(group="Function UDF")
# incompatible with compilation currently
def test_user_def_func_builtin_direct(func, args, expected, benchmark):
    func = UserDefinedFunction(func)

    val = benchmark(func, *args)
    np.testing.assert_allclose(val, expected)

@pytest.mark.composition
@pytest.mark.benchmark(group="UDF as Composition Origin")
def test_udf_composition_origin(comp_mode, benchmark):
    def myFunction(variable, context):
        return [variable[0][1], variable[0][0]]

    myMech = ProcessingMechanism(function=myFunction, input_shapes=3, name='myMech')
    T = TransferMechanism(input_shapes=2, function=Linear)
    c = Composition(pathways=[myMech, T])
    benchmark(c.run, inputs={myMech: [[1, 3, 5]]}, execution_mode=comp_mode)
    np.testing.assert_allclose(c.results[0][0], [3, 1])


@pytest.mark.composition
@pytest.mark.benchmark(group="UDF as Composition Terminal")
def test_udf_composition_terminal(comp_mode, benchmark):
    def myFunction(variable, context):
        return [variable[0][2], variable[0][0]]

    myMech = ProcessingMechanism(function=myFunction, input_shapes=3, name='myMech')
    T2 = TransferMechanism(input_shapes=3, function=Linear)
    c2 = Composition(pathways=[[T2, myMech]])
    benchmark(c2.run, inputs={T2: [[1, 2, 3]]}, execution_mode=comp_mode)
    np.testing.assert_allclose(c2.results[0][0], [3, 1])


def test_udf_with_pnl_func():
    L = Logistic(gain=2)

    def myFunction(variable, context):
        return L(variable) + 2

    U = UserDefinedFunction(custom_function=myFunction, default_variable=[[0, 0, 0]])
    myMech = ProcessingMechanism(function=myFunction, input_shapes=3, name='myMech')
    val1 = myMech.execute(input=[1, 2, 3])
    val2 = U.execute(variable=[[1, 2, 3]])
    np.testing.assert_allclose(val1, val2)
    np.testing.assert_allclose(val1, [L([1, 2, 3]) + 2])


def test_udf_runtime_params_reset():
    def myFunction(variable, x):
        return variable + x

    U = UserDefinedFunction(custom_function=myFunction, x=0)
    assert U.function(0) == 0
    assert U.function(0, params={'x': 1}) == 1
    assert U.function(0) == 0


@pytest.mark.parametrize(
    'expression, parameters, result',
    [
        ('x + y', {'x': 2, 'y': 4}, 6),
        ('(x + y) * z', {'x': 2, 'y': 4, 'z': 2}, 12),
        ('x + f(3)', {'x': 1, 'f': lambda x: x}, 4),
        ('x + f (3)', {'x': 1, 'f': lambda x: x}, 4),
        ('np.sum([int(x), 2])', {'x': 1, 'np': np}, 3),
        (
            '(x * y) / 3 + f(z_0, z) + z0 - (x**y) * VAR',
            {'x': 2, 'y': 3, 'f': lambda a, b: a + b, 'z_0': 1, 'z0': 1, 'z': 1, 'VAR': 1},
            -3
        )
    ]
)
@pytest.mark.parametrize('explicit_udf', [True, False])
def test_expression_execution(expression, parameters, result, explicit_udf):
    if explicit_udf:
        u = UserDefinedFunction(custom_function=expression, **parameters)
    else:
        m = ProcessingMechanism(function=expression, **parameters)
        u = m.function

    for p in parameters:
        assert p in u.cust_fct_params

    assert u.execute() == result


def _function_test_integration(variable, x, y, z):
    return x * y + z


@pytest.mark.parametrize(
    'function',
    [
        (lambda variable, x, y, z: x * y + z),
        'x * y + z',
        _function_test_integration
    ]
)
def test_integration(function):
    u = UserDefinedFunction(
        custom_function=function,
        x=2,
        y=3,
        z=5,
        stateful_parameter='x'
    )

    assert u.execute() == 11
    assert u.execute() == 38


class TestUserDefFunc:
    def test_udf_creates_parameter_ports(self):
        def func(input=[[0], [0]], p=0, q=1):
            return (p + q) * input

        m = ProcessingMechanism(
            default_variable=[[0], [0]],
            function=UserDefinedFunction(func)
        )

        assert len(m.parameter_ports) == 2
        assert 'p' in m.parameter_ports.names
        assert 'q' in m.parameter_ports.names

    @pytest.fixture
    def mech_with_autogenerated_udf(self):
        def func(input=[[0], [0]], p=0, q=1):
            return (p + q) * input

        m = ProcessingMechanism(
            default_variable=[[0], [0]],
            function=func
        )

        return m

    def test_mech_autogenerated_udf_execute(self, mech_with_autogenerated_udf):
        # Test if execute is working with auto-defined udf's
        val1 = mech_with_autogenerated_udf.execute(input=[[1], [1]])
        np.testing.assert_allclose(val1, np.array([[1], [1]]))

    def test_autogenerated_udf(self, mech_with_autogenerated_udf):
        assert isinstance(mech_with_autogenerated_udf.function, UserDefinedFunction)

    def test_autogenerated_udf_creates_parameter_ports(self, mech_with_autogenerated_udf):
        assert len(mech_with_autogenerated_udf.parameter_ports) == 2
        assert 'p' in mech_with_autogenerated_udf.parameter_ports.names
        assert 'q' in mech_with_autogenerated_udf.parameter_ports.names

    def test_autogenerated_udf_creates_parameters(self, mech_with_autogenerated_udf):
        assert hasattr(mech_with_autogenerated_udf.function.parameters, 'p')
        assert hasattr(mech_with_autogenerated_udf.function.parameters, 'q')

        assert mech_with_autogenerated_udf.function.parameters.p.default_value == 0
        assert mech_with_autogenerated_udf.function.parameters.q.default_value == 1

        assert not hasattr(mech_with_autogenerated_udf.function.class_parameters, 'p')
        assert not hasattr(mech_with_autogenerated_udf.function.class_parameters, 'q')

    def test_autogenerated_udf_parameters_states_have_source(self, mech_with_autogenerated_udf):
        for p in mech_with_autogenerated_udf.parameter_ports:
            assert p.source is getattr(mech_with_autogenerated_udf.function.parameters, p.name)



import numpy as np
import pytest

from psyneulink.core.components.functions.function import FunctionError
from psyneulink.core.components.functions.nonstateful.learningfunctions import Hebbian, Reinforcement, TDLearning
from psyneulink.core.components.functions.nonstateful.objectivefunctions import Distance
from psyneulink.core.components.functions.nonstateful.distributionfunctions import NormalDist, ExponentialDist, \
    UniformDist, GammaDist, WaldDist, DriftDiffusionAnalytical
from psyneulink.core.components.functions.stateful.integratorfunctions import SimpleIntegrator, \
    AdaptiveIntegrator, DriftDiffusionIntegrator, OrnsteinUhlenbeckIntegrator, FitzHughNagumoIntegrator, \
    AccumulatorIntegrator, DualAdaptiveIntegrator
from psyneulink.core.components.functions.nonstateful.transferfunctions import Linear, Exponential, Logistic, SoftMax
from psyneulink.core.components.functions.nonstateful.transformfunctions import \
    CombineMeans, LinearCombination, MatrixTransform, Reduce
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.globals.keywords import \
    MAX_ABS_INDICATOR, MAX_ABS_ONE_HOT, MAX_ABS_VAL, MAX_INDICATOR, MAX_ONE_HOT, MAX_VAL, \
    MEAN, MEDIAN, NAME, OWNER_VALUE, PROB, STANDARD_DEVIATION, VARIABLE, VARIANCE

class TestProcessingMechanismFunctions:

    @pytest.mark.benchmark(group="ProcessingMechanism[DefaultFunction]")
    @pytest.mark.parametrize("variable", [[1, 2, 3, 4],
                                          [1., 2., 3., 4.],
                                          np.asarray([1., 2., 3., 4.], dtype=np.int8),
                                          np.asarray([1., 2., 3., 4.], dtype=np.int16),
                                          np.asarray([1., 2., 3., 4.], dtype=np.int32),
                                          np.asarray([1., 2., 3., 4.], dtype=np.int64),
                                          np.asarray([1., 2., 3., 4.], dtype=np.float32),
                                          np.asarray([1., 2., 3., 4.], dtype=np.float64),
                                          [[1, 2, 3, 4]],
                                          [[1., 2., 3., 4.]],
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.int8),
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.int16),
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.int32),
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.int64),
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.float32),
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.float64),
                                         ],
                             ids=["list.int", "list.float", "np.1d.i8", "np.1d.i16", "np.1d.i32", "np.1d.i64", "np.1d.f32", "np.1d.f64",
                                  "list2d.int", "list2d.float", "np.2d.i8", "np.2d.i16", "np.2d.i32", "np.2d.i64", "np.2d.f32", "np.2d.f64",
                             ])
    def test_processing_mechanism_default_function(self, mech_mode, variable, benchmark):
        PM = ProcessingMechanism(default_variable=[0, 0, 0, 0])
        ex = pytest.helpers.get_mech_execution(PM, mech_mode)

        res = benchmark(ex, variable)
        np.testing.assert_allclose(res, [[1., 2., 3., 4.]])

    def test_processing_mechanism_linear_function(self):

        PM1 = ProcessingMechanism()
        PM1.execute(1.0)
        np.testing.assert_allclose(PM1.value, 1.0)

        PM2 = ProcessingMechanism(function=Linear(slope=2.0,
                                                  intercept=1.0))
        PM2.execute(1.0)
        np.testing.assert_allclose(PM2.value, 3.0)

    @pytest.mark.parametrize(
        "function,expected", [
            (LinearCombination, [[1.]]),
            (Reduce, [[1.]]),
            (CombineMeans, [[1.0]]),
            (Exponential, [[2.71828183]]),
            (Logistic, [[0.73105858]]),
            (SoftMax, [[1, ]]),
            (SimpleIntegrator, [[1.]]),
            (AdaptiveIntegrator, [[1.]]),
            (DriftDiffusionIntegrator, [[[1.]], [[1.]]]),
            (OrnsteinUhlenbeckIntegrator, [[[-1.]], [[1.]]]),
            (AccumulatorIntegrator, [[0.]]),
            (FitzHughNagumoIntegrator, [[[0.05127053]], [[0.00279552]], [[0.05]]]),
            (DualAdaptiveIntegrator, [[0.1517455]]),
            (DriftDiffusionAnalytical, [[1.19932930e+00],
                                        [3.35350130e-04],
                                        [1.19932930e+00],
                                        [2.48491374e-01],
                                        [1.48291009e+00],
                                        [1.19932930e+00],
                                        [2.48491374e-01],
                                        [1.48291009e+00]]),
            (NormalDist, [[-0.51529709]]),
            (ExponentialDist, [[0.29964231]]),
            (UniformDist, [[0.25891675]]),
            (GammaDist, [[0.29964231]]),
            (WaldDist, [[0.73955962]]),
        ],
        ids=lambda x: getattr(x, "componentName", ""))
    def test_processing_mechanism_function(self, function, expected):
        PM = ProcessingMechanism(function=function)
        res = PM.execute(1.0)
        np.testing.assert_allclose(res, expected, rtol=1e-5, atol=1e-8)

    # COMMENTED OUT BECAUSE OF MATLAB ENGINE:
    # def test_processing_mechanism_NavarroAndFuss_function(self):
    #     PM1 = ProcessingMechanism(function=NavarroAndFuss)
    #     PM1.execute(1.0)
    #     # np.testing.assert_allclose(PM1.value, 1.0, rtol=1e-5, atol=1e-8)

    def test_processing_mechanism_Distance_function(self):
        PM1 = ProcessingMechanism(function=Distance,
                                  default_variable=[[0,0], [0,0]])
        PM1.execute([[1, 2], [3, 4]])
        # np.testing.assert_allclose(PM1.value, 1.0)

    def test_processing_mechanism_Hebbian_function(self):
        PM1 = ProcessingMechanism(function=Hebbian,
                                  default_variable=[[0.0], [0.0], [0.0]])
        PM1.execute([[1.0], [2.0], [3.0]])
        # np.testing.assert_allclose(PM1.value, 1.0)

    def test_processing_mechanism_Reinforcement_function(self):
        PM1 = ProcessingMechanism(function=Reinforcement,
                                  default_variable=[[0.0], [0.0], [0.0]])
        PM1.execute([[1.0], [2.0], [3.0]])
        # np.testing.assert_allclose(PM1.value, 1.0)

    # COMMENTING OUT BECAUSE BACK PROP FN DOES NOT WORK WITH UNRESTRICTED MECHANISM
    # def test_processing_mechanism_BackPropagation_function(self):
    #     PM1 = ProcessingMechanism(function=BackPropagation,
    #                                          default_variable=[[0.0], [0.0], [0.0]])
    #     PM1.execute([[1.0], [2.0], [3.0]])
    #     PM1.execute(1.0)
    #     # np.testing.assert_allclose(PM1.value, 1.0)

    def test_processing_mechanism_TDLearning_function(self):
        PM1 = ProcessingMechanism(function=TDLearning,
                                  default_variable=[[0.0], [0.0], [0.0]])
        PM1.execute([[1.0], [2.0], [3.0]])
        # np.testing.assert_allclose(PM1.value, 1.0)

    def test_processing_mechanism_multiple_input_ports(self):

        PM1 = ProcessingMechanism(input_shapes=[4, 4], function=LinearCombination, input_ports=['input_1', 'input_2'])
        PM2 = ProcessingMechanism(input_shapes=[2, 2, 2], function=LinearCombination, input_ports=['1', '2', '3'])
        PM1.execute([[1, 2, 3, 4], [5, 4, 2, 2]])
        PM2.execute([[2, 0], [1, 3], [1, 0]])
        np.testing.assert_allclose(PM1.value, [[6, 6, 5, 6]])
        np.testing.assert_allclose(PM2.value, [[4, 3]])

        mech = ProcessingMechanism(name='mech',
                                   input_ports=['a','b','c'])
        result = mech.execute([[1],[2],[3]])
        assert mech.input_ports.names == ['a','b','c']
        assert mech.output_ports.names == ['a','b','c']
        assert result.tolist() == [[1],[2],[3]]     # Note: this is the result of the Mechanism's function
        assert mech.output_values == [[1],[2],[3]]  # Note: this is list of values of its OutputPorts

        mech = ProcessingMechanism(name='mech',
                                   input_ports=['a','b','c'],
                                   output_ports=['d','e'])
        result = mech.execute([[1],[2],[3]])
        assert mech.output_ports.names == ['d','e']
        assert result.tolist() == [[1],[2],[3]]  # Note: this is the result of the Mechanism's function
        assert mech.output_values == [[1],[2]]   # Note: this is list of values of its OutputPorts

        mech = ProcessingMechanism(name='mech',
                                   input_ports=['a','b','c'],
                                   output_ports=[{}, {}])
        result = mech.execute([[1],[2],[3]])
        assert mech.output_ports.names == ['a','b']
        assert result.tolist() == [[1],[2],[3]]  # Note: this is the result of the Mechanism's function
        assert mech.output_values == [[1],[2]]   # Note: this is list of values of its OutputPorts

        mech = ProcessingMechanism(name='mech',
                                   input_ports=['a','b','c'],
                                   output_ports=[OutputPort(), OutputPort()])
        result = mech.execute([[1],[2],[3]])
        assert mech.output_ports.names == ['a','b']
        assert result.tolist() == [[1],[2],[3]]  # Note: this is the result of the Mechanism's function
        assert mech.output_values == [[1],[2]]   # Note: this is list of values of its OutputPorts

        mech = ProcessingMechanism(name='mech',
                                   input_ports=['a','b','c'],
                                   output_ports=[{NAME: 'd'}, {VARIABLE: (OWNER_VALUE, 0)}])
        result = mech.execute([[1],[2],[3]])
        assert mech.output_ports.names == ['d','a']
        assert result.tolist() == [[1],[2],[3]]  # Note: this is the result of the Mechanism's function
        assert mech.output_values == [[1],[1]]   # Note: this is list of values of its OutputPorts

        mech = ProcessingMechanism(name='mech',
                                   input_ports=['a','b','c'],
                                   output_ports=[OutputPort(),
                                                 OutputPort(name='3rd', variable=(OWNER_VALUE, 2))])
        result = mech.execute([[1],[2],[3]])
        assert mech.output_ports.names == ['a','3rd']
        assert result.tolist() == [[1],[2],[3]]  # Note: this is the result of the Mechanism's function
        assert mech.output_values == [[1],[3]]   # Note: this is list of values of its OutputPorts

        mech = ProcessingMechanism(name='mech',
                                   input_ports=['a','b','c'],
                                   output_ports=[OutputPort(name='1st'), OutputPort(variable=(OWNER_VALUE, 2))])
        result = mech.execute([[1],[2],[3]])
        assert mech.output_ports.names == ['1st','c']
        assert result.tolist() == [[1],[2],[3]]  # Note: this is the result of the Mechanism's function
        assert mech.output_values == [[1],[3]]   # Note: this is list of values of its OutputPorts

        mech = ProcessingMechanism(name='mech',
                                   input_ports=['a','b','c'],
                                   output_ports=[OutputPort(name='1st'), OutputPort()])
        result = mech.execute([[1],[2],[3]])
        assert mech.output_ports.names == ['1st','b']
        assert result.tolist() == [[1],[2],[3]]  # Note: this is the result of the Mechanism's function
        assert mech.output_values == [[1],[2]]   # Note: this is list of values of its OutputPorts


class TestMatrixTransformFunction:

    def test_valid_matrix_specs(self):
        # Note: default matrix specification is None

        PM_default = ProcessingMechanism(function=MatrixTransform())
        PM_default.execute(1.0)

        np.testing.assert_allclose(PM_default.value, 1.0)

        PM_default_len_2_var = ProcessingMechanism(function=MatrixTransform(default_variable=[[0.0, 0.0]]),
                                                   default_variable=[[0.0, 0.0]])
        PM_default_len_2_var.execute([[1.0, 2.0]])

        np.testing.assert_allclose(PM_default_len_2_var.value, [[1.0, 2.0]])

        PM_default_2d_var = ProcessingMechanism(function=MatrixTransform(default_variable=[[0.0, 0.0],
                                                                                        [0.0, 0.0],
                                                                                        [0.0, 0.0]]),
                                                   default_variable=[[0.0, 0.0],
                                                                     [0.0, 0.0],
                                                                     [0.0, 0.0]])

        #  [1.0   0.0]           [1.0   0.0]
        #  [0.0   2.0]    *      [0.0   1.0]
        #  [3.0   0.0]
        PM_default_2d_var.execute([[1.0, 0.0],
                                   [0.0, 2.0],
                                   [3.0, 0.0]])

        np.testing.assert_allclose(PM_default_2d_var.value, [[1.0, 0.0],
                                                     [0.0, 2.0],
                                                     [3.0, 0.0]])

        # PM_float = ProcessingMechanism(function=MatrixTransform(matrix=4.0))
        # PM_float.execute(1.0)
        #
        # np.testing.assert_allclose(PM_float.value, 4.0)

        PM_1d_list = ProcessingMechanism(function=MatrixTransform(matrix=[4.0]))
        PM_1d_list.execute(1.0)

        np.testing.assert_allclose(PM_1d_list.value, 4.0)

        PM_2d_list = ProcessingMechanism(function=MatrixTransform(matrix=[[4.0, 5.0],
                                                                       [6.0, 7.0],
                                                                       [8.0, 9.0],
                                                                       [10.0, 11.0]],
                                                               default_variable=[[0.0, 0.0, 0.0, 0.0],
                                                                                 [0.0, 0.0, 0.0, 0.0],
                                                                                 [0.0, 0.0, 0.0, 0.0]]),
                                         default_variable=[[0.0, 0.0, 0.0, 0.0],
                                                           [0.0, 0.0, 0.0, 0.0],
                                                           [0.0, 0.0, 0.0, 0.0]])
        PM_2d_list.execute([[1.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 1.0, 0.0],
                            [0.0, 0.0, 0.0, 1.0]])

        np.testing.assert_allclose(PM_2d_list.value, [[4.0, 5.0],
                                              [8.0, 9.0],
                                              [10.0, 11.0]])

        PM_1d_array = ProcessingMechanism(function=MatrixTransform(matrix=np.array([4.0])))
        PM_1d_array.execute(1.0)

        np.testing.assert_allclose(PM_1d_array.value, 4.0)

        PM_2d_array = ProcessingMechanism(function=MatrixTransform(matrix=np.array([[4.0]])))
        PM_2d_array.execute(1.0)

        np.testing.assert_allclose(PM_2d_array.value, 4.0)

        PM_matrix = ProcessingMechanism(function=MatrixTransform(matrix=np.array([[4.0]])))
        PM_matrix.execute(1.0)

        np.testing.assert_allclose(PM_matrix.value, 4.0)

    def test_invalid_matrix_specs(self):

        with pytest.raises(FunctionError) as error_text:
            PM_mismatched_float = ProcessingMechanism(function=MatrixTransform(default_variable=0.0,
                                                                        matrix=[[1.0, 0.0, 0.0, 0.0],
                                                                                [0.0, 2.0, 0.0, 0.0],
                                                                                [0.0, 0.0, 3.0, 0.0],
                                                                                [0.0, 0.0, 0.0, 4.0]]),
                                                  default_variable=0.0)
        assert "Specification of matrix and/or default_variable" in str(error_text.value) and \
               "not compatible for multiplication" in str(error_text.value)

        with pytest.raises(FunctionError) as error_text:
            PM_mismatched_matrix = ProcessingMechanism(function=MatrixTransform(default_variable=[[0.0, 0.0],
                                                                                          [0.0, 0.0],
                                                                                          [0.0, 0.0]],
                                                                        matrix=[[1.0, 0.0, 0.0, 0.0],
                                                                                [0.0, 2.0, 0.0, 0.0],
                                                                                [0.0, 0.0, 3.0, 0.0],
                                                                                [0.0, 0.0, 0.0, 4.0]]),
                                                  default_variable=[[0.0, 0.0],
                                                                    [0.0, 0.0],
                                                                    [0.0, 0.0]])
        assert "Specification of matrix and/or default_variable" in str(error_text.value) and \
               "not compatible for multiplication" in str(error_text.value)

class TestProcessingMechanismStandardOutputPorts:

    @pytest.mark.benchmark
    @pytest.mark.parametrize("op, expected", [(MAX_ONE_HOT, [[0, 2, 0]]),
                                              (MAX_INDICATOR, [[0, 1, 0]]),
                                              (MAX_ABS_INDICATOR, [[0, 0, 1]]),
                                              (MAX_ABS_ONE_HOT, [[0, 0, 4]]),
                                              (MAX_VAL, [[2]]),
                                              (PROB, [[[0, 2, 0]]]),
                                             ],
                             ids=lambda x: x if isinstance(x, str) else "")
    def test_output_ports(self, mech_mode, op, expected, benchmark):
        benchmark.group = "Output Port Op: {}".format(op)
        PM1 = ProcessingMechanism(default_variable=[0, 0, 0], output_ports=[op])
        var = [1, 2, 4] if op in {MEAN, MEDIAN, STANDARD_DEVIATION, VARIANCE} else [1, 2, -4]
        ex = pytest.helpers.get_mech_execution(PM1, mech_mode)

        res = benchmark(ex, var)
        np.testing.assert_allclose(res, expected)

    # FIXME: These variants don't compile (use UDFs)
    @pytest.mark.parametrize("op, expected", [(MEAN, [2.33333333]),
                                              (MEDIAN, [2]),
                                              (STANDARD_DEVIATION, [1.24721913]),
                                              (VARIANCE, [1.55555556]),
                                              (MAX_ABS_VAL, [4]),
                                             ],
                             ids=lambda x: x if isinstance(x, str) else "")
    def test_output_ports2(self, op, expected):
        PM1 = ProcessingMechanism(default_variable=[0, 0, 0], output_ports=[op])
        var = [1, 2, 4] if op in {MEAN, MEDIAN, STANDARD_DEVIATION, VARIANCE} else [1, 2, -4]
        PM1.execute(var)
        np.testing.assert_allclose(PM1.output_ports[0].value, expected)



import functools
import numpy as np
import psyneulink as pnl
import psyneulink.core.components.functions.nonstateful.transferfunctions

Input_Layer = pnl.TransferMechanism(
    name='Input Layer',
    function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic,
    params={pnl.INPUT_LABELS_DICT:{'red': [-1, 30]}},
    default_variable=np.zeros((2,)))

Hidden_Layer_1 = pnl.TransferMechanism(
    name='Hidden Layer_1',
    function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic(),
    default_variable=np.zeros((5,)))

Hidden_Layer_2 = pnl.TransferMechanism(
    name='Hidden Layer_2',
    function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic(),
    default_variable=[0, 0, 0, 0])

Output_Layer = pnl.TransferMechanism(
    name='Output Layer',
    function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic,
    default_variable=[0, 0, 0])

Input_Weights_matrix = (np.arange(2 * 5).reshape((2, 5)) + 1) / (2 * 5)
Middle_Weights_matrix = (np.arange(5 * 4).reshape((5, 4)) + 1) / (5 * 4)
Output_Weights_matrix = (np.arange(4 * 3).reshape((4, 3)) + 1) / (4 * 3)

# This Projection will be used by the Process below by referencing it in the Process' pathway;
#    note: sender and receiver args don't need to be specified
Input_Weights = pnl.MappingProjection(
    name='Input Weights',
    matrix=Input_Weights_matrix
)

# This Projection will be used by the Process below by assigning its sender and receiver args
#    to mechanisms in the pathway
Middle_Weights = pnl.MappingProjection(
    name='Middle Weights',
    sender=Hidden_Layer_1,
    receiver=Hidden_Layer_2,
    matrix=Middle_Weights_matrix
)

# Treated same as Middle_Weights Projection
Output_Weights = pnl.MappingProjection(
    name='Output Weights',
    sender=Hidden_Layer_2,
    receiver=Output_Layer,
    matrix=Output_Weights_matrix
)

z = pnl.Pathway(
    pathway=[
        Input_Layer,
        # The following reference to Input_Weights is needed to use it in the pathway
        #    since it's sender and receiver args are not specified in its declaration above
        Input_Weights,
        Hidden_Layer_1,
        # Middle_Weights,
        # No Projection specification is needed here since the sender arg for Middle_Weights
        #    is Hidden_Layer_1 and its receiver arg is Hidden_Layer_2
        Hidden_Layer_2,
        # Output_Weights,
        # Output_Weights does not need to be listed for the same reason as Middle_Weights
        # If Middle_Weights and/or Output_Weights is not declared above, then the Process
        #    will assign a default for rhe missing Projection
        Output_Layer
    ]
)


def print_header(comp, context):
    print("\n\n**** Time: ", comp.scheduler.get_clock(context).simple_time)


def show_target(comp):
    i = comp.external_input_values
    t = comp.pathways[0].target.input_ports[0].parameters.value.get(comp)
    print('\nOLD WEIGHTS: \n')
    print('- Input Weights: \n', Input_Weights.parameters.matrix.get(comp))
    print('- Middle Weights: \n', Middle_Weights.parameters.matrix.get(comp))
    print('- Output Weights: \n', Output_Weights.parameters.matrix.get(comp))

    print('\nSTIMULI:\n\n- Input: {}\n- Target: {}\n'.format(i, t))
    print('ACTIVITY FROM OLD WEIGHTS: \n')
    print('- Middle 1: \n', Hidden_Layer_1.parameters.value.get(comp))
    print('- Middle 2: \n', Hidden_Layer_2.parameters.value.get(comp))
    print('- Output:\n', Output_Layer.parameters.value.get(comp))


comp = pnl.Composition(name='Multilayer-Learning',
                       pathways=[(z, pnl.BackPropagation)],
                       targets=[0, 0, 1],
                       learning_rate=2.0,
                       prefs={pnl.VERBOSE_PREF: False,
                              pnl.REPORT_OUTPUT_PREF: True}
)

# Log Middle_Weights of MappingProjection to Hidden_Layer_2
# Hidden_Layer_2.set_log_conditions('Middle Weights')
Middle_Weights.set_log_conditions('mod_matrix')

comp.reportOutputPref = True
# Shows graph will full information:
# comp.show_graph(show_dimensions=pnl.ALL)
comp.show_graph(show_learning=pnl.ALL, show_node_structure=True)
# comp.show_graph(show_learning=pnl.ALL, show_processes=True)
# comp.show_graph(show_learning=pnl.ALL, show_dimensions=pnl.ALL, show_mechanism_structure=True)
# Shows minimal graph:
# comp.show_graph()


stim_list = {Input_Layer: ['red']}
target_list = {Output_Layer: [[0, 0, 1]]}

comp.learn(
    num_trials=1,
    inputs=stim_list,
    targets=target_list,
    clamp_input=pnl.SOFT_CLAMP,
    call_before_trial=functools.partial(print_header, comp),
    call_after_trial=functools.partial(show_target, comp),
    termination_processing={pnl.TimeScale.TRIAL: pnl.AfterNCalls(Output_Layer, 1)},
    animate={'show_learning':pnl.ALL, 'unit':pnl.EXECUTION_SET, pnl.SAVE_IMAGES:True},
)

# Print out logged weights for Middle_Weights
# print('\nMiddle Weights (to Hidden_Layer_2): \n', Hidden_Layer_2.log.nparray(entries='Middle Weights', header=False))
print('\nMiddle Weights (to Hidden_Layer_2): \n', Middle_Weights.log.nparray(entries='mod_matrix', header=False))



from collections.abc import Iterable
import numpy as np
import pytest

from psyneulink.core.globals.utilities import (
    convert_all_elements_to_np_array, extended_array_equal, prune_unused_args, update_array_in_place
)


@pytest.mark.parametrize(
    'arr, expected',
    [
        ([[0], [0, 0]], np.array([np.array([0]), np.array([0, 0])], dtype=object)),
        # should test these but numpy cannot easily create an array from them
        # [np.ones((2,2)), np.zeros((2,1))]
        # [np.array([[0]]), np.array([[[ 1.,  1.,  1.], [ 1.,  1.,  1.]]])]

    ]
)
def test_convert_all_elements_to_np_array(arr, expected):
    converted = convert_all_elements_to_np_array(arr)

    # no current numpy methods can test this
    def check_equality_recursive(arr, expected):
        if (
            not isinstance(arr, Iterable)
            or (isinstance(arr, np.ndarray) and arr.ndim == 0)
        ):
            assert arr == expected
        else:
            assert isinstance(expected, type(arr))
            assert len(arr) == len(expected)

            for i in range(len(arr)):
                check_equality_recursive(arr[i], expected[i])

    check_equality_recursive(converted, expected)


def f():
    pass


def g(a):
    pass


def h(a, b=None):
    pass


def i(b=None):
    pass


@pytest.mark.parametrize(
    'func, args, kwargs, expected_pruned_args, expected_pruned_kwargs', [
        (f, 1, {'a': 1}, [], {}),
        (g, 1, {'x': 1}, [1], {}),
        (g, None, {'a': 1}, [], {'a': 1}),
        (h, None, {'a': 1, 'b': 1, 'c': 1}, [], {'a': 1, 'b': 1}),
        (h, [1, 2, 3], None, [1], {}),
        (i, None, {'a': 1, 'b': 1, 'c': 1}, [], {'b': 1}),
        (i, [1, 2, 3], None, [], {}),
    ]
)
def test_prune_unused_args(func, args, kwargs, expected_pruned_args, expected_pruned_kwargs):
    pruned_args, pruned_kwargs = prune_unused_args(func, args, kwargs)

    assert pruned_args == expected_pruned_args
    assert pruned_kwargs == expected_pruned_kwargs


regular_np_array_parametrization = [
    [],
    0,
    np.array([]),
    [0],
    [[[[[0]]]]],
    [[[[[1]]]]],
    [[1], [0]],
    [[[[[0]]], [[[1]]]]],
]


@pytest.mark.parametrize('a', regular_np_array_parametrization)
@pytest.mark.parametrize('b', regular_np_array_parametrization)
def test_extended_array_equal_regular(a, b):
    assert extended_array_equal(a, b) == np.array_equal(a, b)


irregular_np_array_parametrization = [
    ([0, [1, 0]], [0, [1, 0]], True),
    ([1, [1, 0], [[[1]]]], [1, [1, 0], [[[1]]]], True),
    ([np.array([0, 0]), np.array([0])], [np.array([0, 0]), np.array([0])], True),
    ([1, [], [[[]]]], [1, [], [[[]]]], True),
    ([['ab'], None], [['ab'], None], True),
    ([[0, None, 'ab'], [1, 0]], [[0, None, 'ab'], [1, 0]], True),

    ([0, [0, 0]], [0, [1, 0]], False),
    ([1, [1, 0], [[[1]]]], [], False),
    ([1, [], [[[]]]], [], False),
    ([['ab'], None], [['ab'], 0], False),
    ([[0, None, 'a'], [1, 0]], [[0, None, 'ab'], [1, 0]], False),
]


@pytest.mark.parametrize(
    'a, b, equal', irregular_np_array_parametrization
)
def test_extended_array_equal_irregular(a, b, equal):
    assert extended_array_equal(a, b) == equal


@pytest.mark.parametrize(
    'a',
    [x[0] for x in irregular_np_array_parametrization]
    + [x[1] for x in irregular_np_array_parametrization]
)
def test_extended_array_equal_irregular_identical(a):
    assert extended_array_equal(a, a)


@pytest.mark.parametrize(
    'target, source',
    [
        ([[0, 0], [0, 0]], [[1, 1], [1, 1]]),
        ([[0], [0, 0]], [[1], [1, 1]]),
    ]
)
def test_update_array_in_place(target, source):
    target = convert_all_elements_to_np_array(target)
    source = convert_all_elements_to_np_array(source)
    old_target = target

    update_array_in_place(target, source)

    len_target = len(target)
    assert len_target == len(source)
    assert len_target == len(old_target)
    for i in range(len_target):
        np.testing.assert_array_equal(target[i], source[i])
        np.testing.assert_array_equal(old_target[i], source[i])
        np.testing.assert_array_equal(target[i], old_target[i])


@pytest.mark.parametrize(
    'target, source',
    [
        ([[0], [0, 0]], [[1], [1, 1, 1]]),
        ([0, [0, 0]], [[1], [1, 1]]),
    ]
)
def test_update_array_in_place_failures(target, source):
    target = convert_all_elements_to_np_array(target)
    source = convert_all_elements_to_np_array(source)
    old_target = target

    with pytest.raises(ValueError):
        update_array_in_place(target, source)

    len_target = len(target)
    assert len_target == len(source)
    assert len_target == len(old_target)
    for i in range(len_target):
        assert not np.array_equal(target[i], source[i])
        assert not np.array_equal(old_target[i], source[i])
        np.testing.assert_array_equal(target[i], old_target[i])



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

# ****************************************  EpisodicMemoryMechanism ****************************************************

"""

Contents
--------

  * `EpisodicMemoryMechanism_Overview`
  * `EpisodicMemoryMechanism_Creation`
      - `EpisodicMemoryMechanism_Creation_InputPorts`
      - `EpisodicMemoryMechanism_Creation_Function_Parameters`
      - `EpisodicMemoryMechanism_Creation_OutputPorts`
  * `EpisodicMemoryMechanism_Structure`
      - `EpisodicMemoryMechanism_Memory_Fields`
      - `EpisodicMemoryMechanism_Input`
      - `EpisodicMemoryMechanism_Function`
      - `EpisodicMemoryMechanism_Output`
  * `EpisodicMemoryMechanism_Execution`
  * `EpisodicMemoryMechanism_Examples`
      - `Formatting entries in memory <EpisodicMemoryMechanism_Examples_Default>`
      - `Initializing memory <EpisodicMemoryMechanism_Examples_Memory_Init>`
      - `Naming InputPorts and OutputPorts <EpisodicMemoryMechanism_Examples_Port_Naming>`
  * `EpisodicMemoryMechanism_Class_Reference`

.. _EpisodicMemoryMechanism_Overview:

Overview
--------

An EpisodicMemoryMechanism is a `ProcessingMechanism` that can store and retrieve items from a content addressable
memory;  that is, on each execution it can store an item presented to it as input, and use that to retrieve an item
from its memory based on the content of the input.  The `MemoryFunction` assigned as its `function
<EpisodicMemoryMechanism.function>` determines how items are stored and retrieved. Each memory is a list or array
composed of items referred to as `memory fields <EpisodicMemoryMechanism_Memory_Fields>`, each of which is a list or
array. Memories can have an arbitrary number of fields, and each of those can be of arbitrary shape, however all
memories for a given instance of an  EpisodicMemoryMechanism must have the same shape (number of fields, and shapes
of corresponding fields).  Each `InputPort` of an EpisodicMemoryMechanism provides the input for a corresponding
field of a memory to be stored and used for retrieval.  By default, each `OutputPort` contains the value of a field
of the last retrieved memory although, as with any Mechanism, OutputPorts can be `configured in other ways
<OutputPort_Customization>`. The full set of stored memories can be accessed from the Mechanism's `memory
<EpisodicMemoryMechanism.memory>` attribute, which references its `function's <EpisodicMemoryMechanism.function>`
memory `Parameter <Parameters>`. Other Parameters of its function (e.g., that regulate the probability of storage
and/or retrieval -- see `ContentAddressableMemory`) can be accessed and/or `modulated <ModulatorySignal_Modulation>`
in the standard way for a Mechanism's `function <EpisodicMemoryMechanism.function>`.

At present, EpisodicMemoryMechanism supports the following two MemoryFunctions:

• `ContentAddressableMemory` -- a general form of memory, that stores and retrieves memories as described above;
  any (weighted) combination of its fields can be used for retrieval.

• `DictionaryMemory` -- a more specific form of memory, that has only two fields, for keys and values, that stores
  these as pairs;  retrieval is based on similarity to the key;  this implements a format commonly used by applications
  that use dictionaries as a form of external memory.


.. _EpisodicMemoryMechanism_Creation:

Creating an EpisodicMemoryMechanism
-----------------------------------

.. _EpisodicMemoryMechanism_Creation_InputPorts:

*InputPorts, Entries and Memory Fields*
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

An EpisodicMemoryMechanism is created by calling its constructor with specifications used to implement its
`MemoryFunction` and the shape of entries stored in `memory <EpisodicMemoryMechanism.memory>`.  The latter
is determined by the number of `fields <EpisodicMemoryMechanism_Memory_Fields>` and shape of each in an entry.
These can be specified using any of the following arguments:

  .. _EpisodicMemoryMechanism_Creation_Default_Variable_and_Size:

  * **default_variable** or **input_shapes** -- these are specified in the standard way that the `variable
    <Mechanism_Base.variable>` is specified for any  `Component` (see `default_variable <Component_Variable>`,
    `input_shapes <Component_Size>`, respectively);  the specified value is passed to the constructor for the
    EpisodicMemoryMechanism's `function <EpisodicMemoryMechanism.function>`), which determines the shape of an entry
    in `memory <EpisodicMemoryMechanism.memory>`;  the `memory <EpisodicMemoryMechanism.memory>` itself remains
    empty until the Mechanism is executed and an item is stored.

    .. hint::
       Use **default_variable** rather than **memory** to specify the shape of memory but keep it empty
       until the first entry is stored; note, however, that since retrieval is executed before storage
       (see `EpisodicMemoryMechanism_Execution`), the first execution will return an entry of zeros.

  * **memory** -- specifies a set of entries to be stored in `memory <EpisodicMemoryMechanism.memory>`;  it is passed
    to the constructor for the EpisodicMemoryMechanism's `function <EpisodicMemoryMechanism.function>`) as its
    **initializer** argument (see `initializer <ContentAddressableMemory.initializer>` for an example).

  * **function** -- this can be used to specify a constructor for the `function <EpisodicMemoryMechanism.function>`,
    in which the **default_variable** or **initializer** arguments are used to specify the shape of entries in `memory
    <EpisodicMemoryMechanism.memory>`.  If **default** variable is used, `memory <EpisodicMemoryMechanism.memory>`
    remains empty until the Mechanism is executed and an item is stored. If **initializer** is used, the items
    specified are stored in `memory <EpisodicMemoryMechanism.memory>` and are available for retrieval in the first
    execution of the Mechanism (see `initializer <ContentAddressableMemory.initializer>` for an example).

The above specifications are also used to create the `input_ports <Mechanism_Base.input_ports>` for the Mechanism
in the same way that the `variable <Mechanism_Base.variable>` is used for any Mechanism (see `Mechanism Variable
<Mechanism_Variable_and_InputPorts>` for additional information), with the number of InputPorts created equal to the
number of fields in an entry of `memory <EpisodicMemoryMechanism.memory>`.  Each `input_port
<EpisodicMemoryMechanism.input_ports>` provides the value assigned to a corresponding field
of the entry stored in `memory <EpisodicMemoryMechanism.memory>`, and used to retrieve one similar to it. By default,
`input_port <EpisodicMemoryMechanism.input_ports>` are named *FIELD_n_INPUT*, where "n" is replaced by the index of
each field; however, they can be named explicitly by specifying a list of strings in the **input_ports** argument of
the constructor; the number of these must equal the number of fields specified in **default_variable** or **input_shapes**.

.. _EpisodicMemoryMechanism_Creation_Function_Parameters:

*Function Parameters*
~~~~~~~~~~~~~~~~~~~~~

Parameters that govern the storage and retrieval process are specified as arguments to the `MemoryFunction` specified
in the **function** of the constructor (for example, see `ContentAddressableMemory` for parameters of the default
`function <EpisodicMemoryMechanism.function>`).

.. _EpisodicMemoryMechanism_Creation_OutputPorts:

*OutputPorts*
~~~~~~~~~~~~~

By default, a number of `OutputPorts <OutputPort>` is created equal to the number of InputPorts, each named either
*RETRIEVED_FIELD_n* or *RETRIEVED_<user specified InputPort name>*, that receive the values of the corresponding
fields of a retrieved memory.  OutputPort names can be specified explicitly, by assigning a list of strings to the
**output_ports** argument of the Mechanism's constructor;  however, in that case, or if any other forms of `OutputPort
specification <OutputPort_Forms_of_Specification>` are assigned to **output_ports**, then only the number of OutputPorts
specified are created, which may not match the number of fields of a retrieved memory.

.. _EpisodicMemoryMechanism_Structure:

Structure
---------

.. _EpisodicMemoryMechanism_Memory_Fields:

*Memory Fields*
~~~~~~~~~~~~~~~

Entries in the `memory <EpisodicMemoryMechanism.memory>` of an EpisodicMemoryMechanism are comprised of fields: lists
or 1d arrays within the outer list or array that comprise each entry. An entry can have an arbitrary number of fields,
and fields can be of arbitrary length.  However, all entries must have the same form (i.e., number of fields and shape
of corresponding fields across entries). One InputPort of the EpisodicMemoryMechanism is assigned to each field. Thus,
fields can be used to store different types of information in each field, and to retrieve entries from memory based on
all fields, or a weighted combination of them (as determined by the `MemoryFunction` assigned to `function
<EpisodicMemoryMechanism.function>`;  for example, this can be used to configure the default function,
`ContentAddressableMemory`, as a form of `key-value dictionary <ContentAddressableMemory_Examples_Weighting_Fields>`).

.. _EpisodicMemoryMechanism_Shape:

.. technical_note::

   The shape of an entry in `memory <EpisodicMemoryMechanism.memory>` is determined by the shape of the Mechanism's
   `variable <Mechanism_Base.variable>`. specified in the **default_variable** or **input_shapes** arguments of its constructor
   (see `EpisodicMemoryMechanism_Creation`).  Each item of `variable <Mechanism_Base.variable>` corresponds to a field.
   Both `memory <EpisodicMemoryMechanism.memory>` and all entries are stored in the EpisodicMemoryMechanism's `function
   <EpisodicMemoryMechanism.function>` as np.ndarrays, the dimensionality of which is determined by the shape of an
   entry and its fields. Fields are always stored as 1d arrays; if all fields have the same length (regular), then
   entries are 2d arrays and `memory <EpisodicMemoryMechanism.memory>` is a 3d array.  However, if fields have
   different lengths (`ragged <https://en.wikipedia.org/wiki/Jagged_array>`_) then, although each field is 1d, an
   entry is also 1d (with dtype='object'), and `memory <EpisodicMemoryMechanism.memory>` is 2d (with dtype='object').

.. _EpisodicMemoryMechanism_Input:

*Input*
~~~~~~~

An EpisodicMemoryMechanism has one or more `input_ports <EpisodicMemoryMechanism.input_ports>` that receive the
entry to be stored and that is used to retrieve an existing entry from its memory.  If the Mechanism is assigned
`ContentAddressableMemory` as its `function <EpisodicMemoryMechanism.function>`, then it can have an arbitrary
number of InputPorts, the input to which is assigned to the corresponding `memory field
<EpisodicMemoryMechanism_Memory_Fields>` of that function. By default InputPorts are named *FIELD_n_INPUT* (see
`EpisodicMemoryMechanism_Creation`). If the Mechanism is assigned `DictionaryMemory` as its `function
<EpisodicMemoryMechanism.function>`, then it is assigned at least one InputPort (named *KEY_INPUT* by default),
and optionally a second (named *VALUE_INPUT*) if **default_variable** or **input_shapes** specifies two items; any additional
fields are ignored.

.. _EpisodicMemoryMechanism_Function:

*Function*
~~~~~~~~~~

The default function is `ContentAddressableMemory` that can store entries with an arbitrary number of fields and
shapes, and retrieve them based on a weighted similarity to any combination of those fields.  This can be configured
as a `key-value dictionary <ContentAddressableMemory_Examples_Weighting_Fields>`, or a more specific function dedicated
to that purpose -- `DictionaryMemory` -- can be assigned, in which entries are restricted to be key-value pairs,
and retrieved based on similarity only to the key.  A custom function can also be specified, so long as it
meets the following requirements:

    * it must accept a list or array as its first argument, the items of which are lists or arrays;

    * it must return a list or array of the same size and shape as the input;

    * it must implement a ``memory`` attribute, that can be accessed by the EpisodicMemoryMechanism's `memory
      <EpisodicMemoryMechanism.memory>` attribute.

    .. technical_note::
       if the function is classed, and has a classmethod _enforce_memory_shape(), this is used to ensure that
       any specification of the **memory** argument in the EpisodicMemoryMechanism's constructor conforms to the
       format required for the memory attribute of the function.

.. _EpisodicMemoryMechanism_Output:

*Output*
~~~~~~~~

By default, an EpisodicMemoryMechanism has a number of `OutputPorts <OutputPort>` equal to its number of InputPorts,
each of which is assigned the value of a corresponding field of the entry retrieved from memory (see
`EpisodicMemoryMechanism_Creation` for naming). However, if OutputPorts were specified in the constructor, then there
may be a different number of OutputPorts than InputPorts and `memory fields <EpisodicMemoryMechanism_Memory_Fields>`,
and thus some of the latter may not be reflected in any of the Mechanism's `output_ports
<EpisodicMemoryMechanism.output_ports>`. If the `function <EpisodicMemoryMechanism.function>` is a `DictionaryMemory`,
then it will have at least one OutputPort, named *KEY_OUTPUT*, that is assigned the key (first field) of the entry
retrieved from memory and, if two fields are specified in **default_variable** or **sze**, the Mechanism will have a
second OutputPort named *VALUE_OUTPUT* that is assigned the value (second field) of the entry retrieved from memory;
any additional ones are ignored and no other OutputPorts are created.

.. _EpisodicMemoryMechanism_Execution:

Execution
---------

When an EpisodicMemoryMechanism is executed, its `function <EpisodicMemoryMechanism.function>` carries out
the following operations:

    * retrieve an item from `memory <EpisodicMemoryMechanism.memory>` based on the `value <InputPort.value>` of
      its `input_ports <EpisodicMemoryMechanism.input_ports>`; if no retrieval is made, then an appropriately
      shaped zero-valued array is returned.
    ..
    * store the `value <InputPort.value>` of its `input_ports <EpisodicMemoryMechanism.input_ports>` as an entry in
      `memory <EpisodicMemoryMechanism.function>`.
    ..
    * assign the value of the entry retrieved to its `output_ports <EpisodicMemoryMechanism.output_ports>`,
      based on how the latter are configured (see `EpisodicMemoryMechanism_Creation_OutputPorts`).

    .. note::
         In general, retrieval is executed before storage, so that the current items is not also retrieved;
         however, the order of storage and retrieval is determined by the EpisodicMemoryMechanism's
         `function <EpisodicMemoryMechanism.function>`.

.. _EpisodicMemoryMechanism_Examples:

Examples
--------

(See `ContentAddressableMemory <ContentAddressableMemory_Examples>` for additional examples of how to use that
Function, including how it can be configured to implement a `key-value dictionary
<ContentAddressableMemory_Examples_Weighting_Fields>`.)

.. _EpisodicMemoryMechanism_Examples_Default:

*Default EpisodicMemoryMechanism*
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The following example creates a default EpisodicMemoryMechanism (with no initial memory)::

    >>> my_em = EpisodicMemoryMechanism()
    >>> my_em.execute([[1,2]])
    array([[0, 0]])
    >>> my_em.execute([[2,5]])
    array([[1., 2.]])

The `default_variable <EpisodicMemoryMechanism_Default_Variable>` for an EpisodicMemoryMechanism is ``[[0,0]]``, so the
format of an entry in `memory <EpisodicMemoryMechanism.memory>` is a single field with two elements. Note that, since
it was not assigned any initial memory, the first execution returns an entry comprised of zeros.  However, the input to
the Mechanism in that execution (``[[1,2]]``) is stored as an entry in `memory <EpisodicMemoryMechanism.memory>`, and on
the second execution, since that is now the only entry in `memory <EpisodicMemoryMechanism.memory>`, that is what is
returned.

.. _EpisodicMemoryMechanism_Examples_Default_Variable:

*Format entries using* **default_variable**
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

In this example, the **default_variable** argument is used to format the entries of `memory
<EpisodicMemoryMechanism.memory>` to have two fields, one with two elements and the other with three::

    >>> my_em = EpisodicMemoryMechanism(default_variable=[[0,0],[0,0,0]])
    >>> my_em.execute([[1,2],[3,4,5]])
    array([array([0, 0]), array([0, 0, 0])], dtype=object)

As in the previous example, the first execution returns zeros since `memory <EpisodicMemoryMechanism.memory>` as not
been initialized;  however, notice that in this case they are formated as specified in **default_variable**.  Note
also that even though a list is specified for **default_variable**, the entry returned is an array; `memory
<EpisodicMemoryMechanism.memory>` and all of its entries are always formated as arrays.

.. _EpisodicMemoryMechanism_Examples_Size:

*Format entries using* **input_shapes**
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The **input_shapes** argument can also be used to format entries::

    >>> my_em = EpisodicMemoryMechanism(input_shapes=[2,3])
    >>> my_em.execute([[1,2],[3,4,5]])
    array([array([0, 0]), array([0, 0, 0])], dtype=object)

Note that each element of **input_shapes** specifies the length of a field
(see `EpisodicMemoryMechanism_Creation_Default_Variable_and_Size` for additional details).

.. _EpisodicMemoryMechanism_Examples_Memory_Init:

*Initialize memory*
~~~~~~~~~~~~~~~~~~~

The **memory** argument of an EpisodicMemoryMechanism's constructor can be used to initialize its `memory
<EpisodicMemoryMechanism.memory>`::

    >>> my_em = EpisodicMemoryMechanism(memory=[[[1,2],[3,4,5]],
    ...                                         [[10,9],[8,7,6]]])
    >>> my_em.execute([[1,2],[3,4,6]])
    array([array([1., 2.]), array([3., 4., 5.])], dtype=object)
    >>> my_em.execute([[1,2],[3,4,6]])
    array([array([1., 2.]), array([3., 4., 6.])], dtype=object)

Note that there was no need to use **default_variable** or **input_shapes** to format entries here, since that is determined
by the entries in the **memory** argument.  If  **default_variable** or **input_shapes** is specified, its shape must be the
same as the entries specified in **memory**.  In this example, since `memory <EpisodicMemoryMechanism.memory>` was
initialized, the first execution returns the closest value to the input, which is used as the retrieval cue.  In the
second execution, the input from the first execution is returned, since it was stored after the first retrieval. The
current contents of memory can be inspected using the `memory <EpisodicMemoryMechanism.memory>` attribute::

    >>> my_em.memory
    array([[array([1., 2.]), array([3., 4., 5.])],
           [array([10.,  9.]), array([8., 7., 6.])],
           [array([1., 2.]), array([3., 4., 6.])]], dtype=object)

Notice that there is only one entry for ``[array([1., 2.]), array([3., 4., 6.])]``, even though it was provided
as input to execute twice.  This is because the default `function <EpisodicMemoryMechanism.function>` is
`ContentAddressableMemory`, and the default value of its `duplicate_entries_allowed
<ContentAddressableMemory.duplicate_entries_allowed>` attribute is False. Notice also that that the dtype of the
`memory <EpisodicMemoryMechanism.memory>` array is object, since its entries are ragged arrays (i.e., ones with fields
of different sizes).

.. _EpisodicMemoryMechanism_Examples_Memory_Init_Function:

*Initialize memory in function*
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The contents of `memory <EpisodicMemoryMechanism.memory>` can also be initialized using the **initializer** argument
in the constructor for the EpisodicMemoryMechanism's `function <EpisodicMemoryMechanism.function>`::

    >>> my_em = EpisodicMemoryMechanism(
    ...                        function=ContentAddressableMemory(initializer=[[[1,2],[3,4,5]],
    ...                                                                      [[10,9],[8,7,6]]]))
    >>> my_em.function.memory
    array([[array([1., 2.]), array([3., 4., 5.])],
           [array([10.,  9.]), array([8., 7., 6.])]], dtype=object)
    >>> my_em.memory
    array([[array([1., 2.]), array([3., 4., 5.])],
           [array([10.,  9.]), array([8., 7., 6.])]], dtype=object)

Notice `memory <EpisodicMemoryMechanism.memory>` actually refers to the contents of the `function
<EpisodicMemoryMechanism.function>`'s `memory <ContentAddressableMemory.memory>` attribute.

.. _EpisodicMemoryMechanism_Examples_Port_Naming:

The `input_ports <EpisodicMemoryMechanism.input_ports>` of an EpisodicMemoryMechanims correspond to fields of entries
in `memory <ContentAddressableMemory.memory>` (see `EpisodicMemoryMechanism_Input`), that by default are named
``FIELD_n_INPUT``::

    >>> my_em.input_ports.names
    ['FIELD_0_INPUT', 'FIELD_1_INPUT']

By default, an EpisodicMemoryMechanims also has the same number of `output_ports <EpisodicMemoryMechanism.output_ports>`
as `input_ports <EpisodicMemoryMechanism.input_ports>`, named correspondingly ``RETRIEVED_FIELD_n``::

    >>> my_em.output_ports.names
    ['RETRIEVED_FIELD_0', 'RETRIEVED_FIELD_1']

These are assigned the values of the fields of the entry retrieved from `memory <ContentAddressableMemory.memory>`.

The names of `input_ports <EpisodicMemoryMechanism.input_ports>` can be customized by specifying a list of names in
the **input_ports** argument of the Mechanism's constructor::

    >>> my_em = EpisodicMemoryMechanism(input_shapes=[2,2,2],
    ...                                 input_ports=['KEY', 'VALUE', 'LABEL'])
    >>> my_em.input_ports.names
    ['KEY', 'VALUE', 'LABEL']

The number of names must be equal to the number of fields in an entry (in this case, 3).  Similarly, the `output_ports
<EpisodicMemoryMechanism.output_ports>` can be named in the **output_ports** argument of the constructor.  In this case,
there can be fewer items specified, in which case the number of fields assigned to OutputPorts will be limited by the
number of OutputPorts specified in the argument::

    >>> my_em = EpisodicMemoryMechanism(memory=[[[1,2],[3,4,5],[6,7]],
    ...                                         [[7,6],[5,4,3],[2,1]]],
    ...                                 input_ports=['KEY', 'VALUE', 'LABEL'],
    ...                                 output_ports=['VALUE_RETRIEVED', 'LABEL_RETRIEVED'])
    >>> my_em.execute([[1,2],[3,4,5],[6,7]])
    array([array([1., 2.]), array([3., 4., 5.]), array([6., 7.])],
          dtype=object)
    >>> my_em.output_ports
    [(OutputPort VALUE_RETRIEVED), (OutputPort LABEL_RETRIEVED)]
    >>> my_em.output_ports.values
    [array([1., 2.]), array([3., 4., 5.])]

Notice that the first two fields of the retrieved entry are assigned to the two OutputPorts, and the third is not
assigned to an OutputPort (see `OutputPort_Customization` for additional information about customizing OutputPorts).

.. _EpisodicMemoryMechanism_Class_Reference:

Class Reference
---------------


"""
import copy
import warnings
from psyneulink._typing import Optional, Union

import numpy as np

from psyneulink.core.components.functions.function import Function
from psyneulink.core.components.functions.stateful.memoryfunctions import \
    DictionaryMemory, ContentAddressableMemory
from psyneulink.core.components.mechanisms.mechanism import MechanismError
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism_Base
from psyneulink.core.components.ports.inputport import InputPort
from psyneulink.core.globals.keywords import EPISODIC_MEMORY_MECHANISM,MULTIPLICATIVE_PARAM, NAME, OWNER_VALUE, VARIABLE
from psyneulink.core.globals.parameters import FunctionParameter, Parameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet
from psyneulink.core.globals.utilities import deprecation_warning, convert_all_elements_to_np_array

__all__ = ['EpisodicMemoryMechanism', 'KEY_INPUT', 'VALUE_INPUT', 'KEY_OUTPUT', 'VALUE_OUTPUT']

KEY_INPUT = 'KEY_INPUT'
VALUE_INPUT = 'VALUE_INPUT'
KEY_OUTPUT = 'KEY_OUTPUT'
VALUE_OUTPUT = 'VALUE_OUTPUT'
DEFAULT_INPUT_PORT_NAME_PREFIX = 'FIELD_'
DEFAULT_INPUT_PORT_NAME_SUFFIX = '_INPUT'
DEFAULT_OUTPUT_PORT_PREFIX = 'RETRIEVED_'


class EpisodicMemoryMechanismError(MechanismError):
    pass


class EpisodicMemoryMechanism(ProcessingMechanism_Base):
    """
    EpisodicMemoryMechanism()

    Subclass of `ProcessingMechanism <ProcessingMechanism>` that implements a content addressable dictionary.
    See `Mechanism <Mechanism_Class_Reference>` for arguments and attributes.

    Arguments
    ---------

    .. _EpisodicMemoryMechanism_Default_Variable:

    default_variable : list or ndarray
        specifies the format used for entries in `memory <EpisodicMemoryMechanism.memory>`.

    memory : list or ndarray
        initial set of entries for `memory <EpisodicMemory.memory>`.  It should be either a 3d regular
        array or a 2d ragged array if the fields of an entry have different lengths; assigned as the
        **initializer** argument of the constructor for the `MemoryFunction` specified in **function**
        (see `initializer <ContentAddressableMemory.initializer>` for default assignment).

    Attributes
    ----------

    input_ports : ContentAddressableList[str, InputPort]
        a list of the Mechanism's `InputPorts <Mechanism_InputPorts>`, the number of which is equal to the
        number of `memory fields <EpisodicMemoryMechanism_Memory_Fields>` in an entry of the Mechanism's `memory
        <EpisodicMemoryMechanism.memory>`, and that are named *FIELD_n_INPUT* by default (see
        `EpisodicMemoryMechanism_Input` and Mechanism `input_ports <Mechanism_Base.input_ports>` for additional
        information).

    function : MemoryFunction
        takes the Mechanism's `variable <Mechanism_Base.variable>` and uses it as a cue to retrieve an entry from
        `memory <EpisodicMemoryMechanism_Function>` based on its distance from existing entries, and then stores
        `variable <Mechanism_Base.variable>` in `memory <EpisodicMemoryMechanism_Function>` (see
        `EpisodicMemoryMechanism_Function` and Mechanism `function <Mechanism_Base.function>` for
        additional information).

    memory : 3d array
        contains entries stored in the `function <EpisodicMemoryMechanism.function>`\\s ``memory`` attribute
        (for example, `memory <ContentAddressableMemory.memory>`).

    output_ports : ContentAddressableList[str, OutputPort]
        a list of the Mechanism's `OutputPorts <Mechanism_OutputPorts>`, the number of which is, by default, equal to
        the number of `memory fields <EpisodicMemoryMechanism_Memory_Fields>` in an entry of the Mechanism's `memory
        <EpisodicMemoryMechanism.memory>`, and each of which is assigned the value of the corresponding field of the
        last entry retrieved from `memory <EpisodicMemoryMechanism.memory>`.  However, as with any Mechanism, the
        number and value of OutputPorts can be customized (see `EpisodicMemoryMechanism_Output` and Mechanism
        `output_ports <Mechanism_Base.output_ports>` for additional information).
    """

    componentName = EPISODIC_MEMORY_MECHANISM

    class Parameters(ProcessingMechanism_Base.Parameters):
        """
            Attributes
            ----------

                variable
                    see `variable <EpisodicMemoryMechanism.variable>`

                    :default value: [[0,0]]
                    :type: ``list``

                function
                    see `function <EpisodicMemoryMechanism.function>`

                    :default value: `DictionaryMemory`
                    :type: `Function`

        """
        variable = Parameter([[0,0]], pnl_internal=True, constructor_argument='default_variable')
        function = Parameter(ContentAddressableMemory, stateful=False, loggable=False)
        memory = FunctionParameter(None, function_parameter_name='initializer')

        def _parse_memory(self, memory):
            if memory is None:
                return memory

            return ContentAddressableMemory._enforce_memory_shape(memory)

    @check_user_specified
    def __init__(self,
                 default_variable:Union[int, list, np.ndarray]=None,
                 input_shapes:Optional[Union[int, list, np.ndarray]]=None,
                 memory:Optional[Union[list, np.ndarray]]=None,
                 function:Optional[Function]=None,
                 params=None,
                 name=None,
                 prefs:   Optional[ValidPrefSet] = None,
                 **kwargs):

        # LEGACY SUPPORT FOR DictionaryMemory
        self._dictionary_memory = (isinstance(function, DictionaryMemory)
                                  or (isinstance(function, type)
                                      and function.__name__ is DictionaryMemory.__name__))
        if self._dictionary_memory:
            # Identify and warn about any deprecated args, and return their values for reassignment
            deprecated_arg_values = deprecation_warning(self, kwargs, {'content_size':'input_shapes'})
            # Assign value of deprecated args to current ones
            if 'input_shapes' in deprecated_arg_values:
                input_shapes = deprecated_arg_values['input_shapes']
            # Need to handle assoc_size specially, since it needs to be added to what was content_size
            if 'assoc_size' in kwargs:
                if isinstance(input_shapes, int):
                    input_shapes = [input_shapes, kwargs['assoc_size']]
                else:
                    input_shapes += kwargs['assoc_size']
                kwargs.pop('assoc_size')

        super().__init__(
            default_variable=default_variable,
            input_shapes=input_shapes,
            function=function,
            params=params,
            name=name,
            prefs=prefs,
            memory=memory,
            **kwargs
        )

    def _handle_default_variable(self, default_variable=None, input_shapes=None, input_ports=None, function=None, params=None):
        """Override to initialize or validate default_variable based on _memory_init or function.memory
            - if memory argument for Mechanism is specified and default_variable is not, use former to specify latter;
            - if both are specified, validate that they are the same shape;
            - if function.memory is specified and default_variable is not, use former to specify latter;
            - if both are specified, validate that they are the same shape;
            - if default_variable is specified and neither memory arg of Mechanism nor function.memory is specified,
                 use default_variable to specify function.memory.
        Note: handling this here insures that input_ports are specified/validated using correct default_variable
        """

        variable_shape = convert_all_elements_to_np_array(default_variable).shape \
                             if default_variable is not None else None
        function_instance = self.function if isinstance(self.function, Function) else None

        # **memory** arg is specified in constructor, so use that to initialize or validate default_variable
        if self.parameters.memory._user_specified:
            memory = self.defaults.memory

            if default_variable is None:
                default_variable = copy.deepcopy(memory[0])
            else:
                entry_shape = convert_all_elements_to_np_array(memory[0]).shape
                if entry_shape != variable_shape:
                    raise EpisodicMemoryMechanismError(f"Shape of 'variable' for {self.name} ({variable_shape}) "
                                                       f"does not match the shape of entries ({entry_shape}) in "
                                                       f"specification of its 'memory' argument.")

        # otherwise, if function.memory is specified, use that to initialize or validate default_variable
        elif function_instance and len(self.function.memory):
            if default_variable is None:
                default_variable = self.function.memory[0]
            else:
                entry_shape = self.function.memory[0].shape
                if entry_shape != variable_shape:
                    raise EpisodicMemoryMechanismError(f"Shape of 'variable' for {self.name} ({variable_shape}) "
                                                       f"does not match the shape of entries ({entry_shape}) in "
                                                       f"the memory of its function ({self.function.name}).")

        return super()._handle_default_variable(default_variable, input_shapes, input_ports, function, params)

    def _instantiate_input_ports(self, context=None):
        """Override to assign default names to input_ports"""

        # IMPLEMENTATION NOTE: REMOVE FIRST CONDITIONAL (LEAVING ELSE CLAUSE) WHEN DictionaryMemory IS RETIRED
        if self._dictionary_memory:
            input_ports = [KEY_INPUT]
            if len(self.parameters.variable.default_value) == 2:
                input_ports.append(VALUE_INPUT)
        else:
            input_ports = self.input_ports or [f'{DEFAULT_INPUT_PORT_NAME_PREFIX}{i}{DEFAULT_INPUT_PORT_NAME_SUFFIX}'
                                               for i in range(len(self.parameters.variable.default_value))]

        super()._instantiate_input_ports(input_ports=input_ports, context=context)

    def _instantiate_function(self, function, function_params, context):
        """Assign memory to function if specified in Mechanism's constructor"""
        memory = self.parameters.memory._get(context)
        if memory is not None:
            function.reset(memory)
        super()._instantiate_function(function, function_params, context)

    def _instantiate_output_ports(self, context=None):
        """Generate OutputPorts with names specified and values with shapes equal to corresponding InputPorts

        If OutputPorts have not been specified, use InputPort names with prefix replaced and suffix removed.

        If any OutputPorts are specified as strings, those are used as names and are the only OutputPorts instantiated
            (even if the number is less than the number of InputPorts)

        If any OutputPorts are specified in a form other than a string, then this method is ignored and OutputPorts
            are instantiated in call to super().instantiate_output_ports;
            note:  in that case, the shapes of the value are as specified and may not necessarily correspond to the
                   shapes of the corresponding Inputs (i.e., memory fields).
        """

        # IMPLEMENTATION NOTE: REMOVE FIRST CONDITIONAL (LEAVING ELSE CLAUSE) WHEN DictionaryMemory IS RETIRED
        if self._dictionary_memory:
            output_ports = [{NAME: KEY_OUTPUT,
                            VARIABLE: (OWNER_VALUE, 0)}]
            if len(self.parameters.variable.default_value) == 2:
                output_ports.append({NAME: VALUE_OUTPUT,
                                     VARIABLE: (OWNER_VALUE, 1)})
            self.parameters.output_ports._set(output_ports, override=True, context=context)

        else:
            output_ports_spec = self.output_ports
            # If output_ports was not specified or they are all strings, instantiate
            # (note: need to instantiate here if output_port specs are strings, to be sure values are correct,
            #    as this doesn't seem to be handled properly by super()._instantiate_output_ports)
            if output_ports_spec is None or all(isinstance(o, str) for o in output_ports_spec):
                output_ports = []
                input_suffix_len = len(DEFAULT_INPUT_PORT_NAME_SUFFIX)

                # Total number should be either the number of names provided, or default to length of Mech's value:
                num_output_ports = len(output_ports_spec) if output_ports_spec else len(self.value)

                for i in range(num_output_ports):

                    # Names specified, so use those:
                    if output_ports_spec:
                        name = self.output_ports[i]
                    # Otherwise, use InputPort names as base, removing DEFAULT_INPUT_PORT_NAME_SUFFIX
                    else:
                        input_port_name = self.input_ports[i].name
                        if not self.parameters.input_ports._user_specified:
                            input_port_name = input_port_name[:-input_suffix_len]
                        name = DEFAULT_OUTPUT_PORT_PREFIX + input_port_name

                    # Add port specification dictionary to list of output_ports
                    output_ports.append({NAME: name,
                                         VARIABLE: (OWNER_VALUE, i)})

                self.parameters.output_ports._set(output_ports, override=True, context=context)

        super()._instantiate_output_ports(context=context)

        # Suppress warning if no Projections from one of the OutputPorts (common case, e.g. for key fields)
        for output_port in self.output_ports:
            output_port.parameters.require_projection_in_composition._set(False, override=True, context=context)

    # IMPLEMENTATION NOTE: REMOVE THIS METHOD WHEN DictionaryMemory IS RETIRED
    def _parse_function_variable(self, variable, context=None):

        if self._dictionary_memory:
            # If assoc has not been specified, add empty list to call to function (which expects two items in its variable)
            if len(variable) != 2:
                return convert_all_elements_to_np_array([variable[0], []])
            else:
                # Check that both are assigned inputs:
                missing_inputs = [self.input_ports.names[i] for i,t in enumerate([v for v in variable]) if t is None]
                if missing_inputs:
                    if len(missing_inputs) == 1:
                        missing_str = 'an input'
                        s = ''
                    else:
                        missing_str = 'inputs'
                        s = 's'
                    raise EpisodicMemoryMechanismError(f"{self.name} is missing {missing_str} for its"
                                                       f" {'and '.join(missing_inputs)} {InputPort.__name__}{s}.")
        return variable

    # IMPLEMENTATION NOTE: REMOVE THIS METHOD WHEN DictionaryMemory IS RETIRED
    def _execute(self, variable=None, context=None, runtime_params=None):

        value = super()._execute(variable=variable,
                                 context=context,
                                 runtime_params=runtime_params,
                                )
        if self._dictionary_memory:
            # Only return content if assoc has not been specified (in which case second element of value should be empty)
            if len(value[1]) == 0:
                return np.delete(value,1)
        return value

    @property
    def memory(self):
        """Return function's memory attribute"""
        try:
            return self.function.memory
        except:
            raise EpisodicMemoryMechanismError(f"Function of {self.name} ({self.function.name}) "
                                               f"has no `memory attribute")
